#!/usr/bin/env python3
# coding=utf-8


"""
Mappers for python stdlog of

IO-streams -> log-level-format

refer ``logician.stdlog.formatters.StdLogLevelFmt`` for log-level-format.
"""

from abc import abstractmethod
from typing import Protocol, IO, override

from logician.format_mappers import StreamFormatMapperComputer
from logician.stdlog.formatters import (
    StdLogLevelFmt,
    StdLogAllLevelSameFmt,
    StdLogAllLevelDiffFmt,
    STDERR_ALL_LVL_SAME_FMT,
    STDERR_ALL_LVL_DIFF_FMT,
    stderr_all_lvl_same_fmt,
)
from logician.stdlog.constants import LOG_LVL as L, LOG_FMT as F


class StdStreamFormatMapperComputer(StreamFormatMapperComputer[L, F], Protocol):
    """
    Interface for the strategies that can compute and then generate mappings of stream -> level-format-mapper for
    python stdlog.

    level-format-map - see ``logician.stdlog.formatters.StdLogLevelFmt``.
    """

    @override
    @abstractmethod
    def compute(
        self, same_fmt_per_lvl: F | bool | None, stream_set: set[IO] | None
    ) -> dict[IO, StdLogLevelFmt]:
        pass  # pragma: no cover


class StdStrFmtMprComputer(StdStreamFormatMapperComputer):
    """
    Implementation of interface ``StdStreamFormatMapperComputer``.

    >>> import sys
    >>> sut = StdStrFmtMprComputer()

    Follows certain rules:

      * Empty ``stream_set``, like, ``{}`` returns an empty dict:

        >>> assert sut.compute(True, set()) == {}

      * ``same_fmt_per_level`` can set same format for all logging levels.

        >>> ret_dict = sut.compute(True, {sys.stderr})
        >>> assert isinstance(ret_dict[sys.stderr], StdLogAllLevelSameFmt)

      * ``same_fmt_per_level`` can enforce same format for all logging levels.

        >>> ret_dict = sut.compute("%(name)s", {sys.stderr})
        >>> assert isinstance(ret_dict[sys.stderr], StdLogAllLevelSameFmt)
        >>> assert ret_dict[sys.stderr].fmt(10) == "%(name)s"   # any level has same formats

      * ``False`` ``same_fmt_per_level`` results in different logging formats per level.

        >>> ret_dict = sut.compute(False, {sys.stdout, sys.stderr})
        >>> assert isinstance(ret_dict[sys.stderr], StdLogAllLevelDiffFmt)
        >>> assert isinstance(ret_dict[sys.stderr], StdLogAllLevelDiffFmt)

      * ``None`` ``stream_set`` and ``same_fmt_per_level`` enforces different-format-per-log-level on the stderr stream.

        >>> ret_dict = sut.compute(None, None)
        >>> assert len(ret_dict) == 1
        >>> assert sys.stderr in ret_dict   # only one entry of stderr stream present
        >>> assert isinstance(ret_dict[sys.stderr], StdLogAllLevelDiffFmt)
    """

    @override
    def compute(
        self, same_fmt_per_lvl: F | bool | None, stream_set: set[IO] | None
    ) -> dict[IO, StdLogLevelFmt]:
        if stream_set is not None:  # accepts empty stream_set
            if same_fmt_per_lvl:
                if isinstance(same_fmt_per_lvl, F):
                    return {
                        stream: StdLogAllLevelSameFmt(same_fmt_per_lvl)
                        for stream in stream_set
                    }
                return {stream: StdLogAllLevelSameFmt() for stream in stream_set}
            return {stream: StdLogAllLevelDiffFmt() for stream in stream_set}
        else:
            if same_fmt_per_lvl:
                if isinstance(same_fmt_per_lvl, F):
                    return stderr_all_lvl_same_fmt(same_fmt_per_lvl)
                return STDERR_ALL_LVL_SAME_FMT
            return STDERR_ALL_LVL_DIFF_FMT



#!/usr/bin/env python3
# coding=utf-8

"""
CLI to

Extract and showcase details about a program's logger configurators.
"""

from logician.main import main_cli

if __name__ == "__main__":
    main_cli()



#!/usr/bin/env python3
# coding=utf-8


"""
Logger interfaces for Logger formatters.
"""

from abc import abstractmethod
from typing import Protocol


class LogLevelFmt[L, F](Protocol):
    """
    Interface to specify how log levels should impact the logging formats.

    For e.g.::

        ERROR: an error occurred.
        logger.name: INFO: some information
        logger.name: DEBUG: [filename.py - func()]: some debug info
        2025-04-03 20:59:39,418: TRACE: [filename.py:218 - func()]: some trace info

    see ``DiffLevelDiffFmt`` for behavior on different log format for different logging levels.

    or the format remains same throughout all logging levels, see ``AllLevelSameFmt``.

    L - type of ``level`` of the logger, e.g. ``int`` for python std log.

    F – type of the returned format, e.g. ``str`` for python std log.
    """

    @abstractmethod
    def fmt(self, level: L) -> F:
        """
        :param level: ``level`` for which the log ``format`` is to be queried.
        :return: format for the queried ``level``.
        """
        pass  # pragma: no cover


class AllLevelSameFmt[L, F](LogLevelFmt[L, F], Protocol):
    """
    Interface to specify how different log levels infer the same logging formats.

    For e.g.::

        - least verbose ERROR level.
        ERROR: an error occurred.

        - less verbose INFO level.
        INFO: some information

        - verbose DEBUG level.
        DEBUG: some debug info

        - most verbose TRACE level.
        TRACE: some trace info
    """

    pass


class DiffLevelDiffFmt[L, F](LogLevelFmt[L, F], Protocol):
    """
    Interface to specify how different log levels should impact the logging formats.

    For e.g.::

        - least verbose ERROR level.
        ERROR: an error occurred.

        - less verbose INFO level.
        logger.name: INFO: some information

        - verbose DEBUG level.
        logger.name: DEBUG: [filename.py - func()]: some debug info

        - most verbose TRACE level.
        2025-04-03 20:59:39,418: TRACE: [filename.py:218 - func()]: some trace info
    """

    @abstractmethod
    def next_approx_level(self, missing_level: L) -> L:
        """
        :param missing_level: A level that was not registered in the logger.
        :return: next approx level if a ``missing_level`` is queried which wasn't already registered in the logger.
        """
        pass  # pragma: no cover



#!/usr/bin/env python3
# coding=utf-8

"""
Logging base interfaces are for implementation as well as extension.
"""

from abc import abstractmethod
from typing import Protocol


class LogLogProtocol[L](Protocol):
    """
    Protocol supporting the log method.

    L - Level type, for e.g. ``int`` for python std logging.
    """

    @abstractmethod
    def log(self, level: L, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class TraceLogProtocol(Protocol):
    """
    Protocol supporting the trace method.
    """

    @abstractmethod
    def trace(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class DebugLogProtocol(Protocol):
    """
    Protocol supporting the debug method.
    """

    @abstractmethod
    def debug(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class InfoLogProtocol(Protocol):
    """
    Protocol supporting the info method.
    """

    @abstractmethod
    def info(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class SuccessLogProtocol(Protocol):
    """
    Protocol supporting the success method.
    """

    @abstractmethod
    def success(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class NoticeLogProtocol(Protocol):
    """
    Protocol supporting the notice method.
    """

    @abstractmethod
    def notice(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class CommandLogProtocol(Protocol):
    """
    Protocol supporting the command logging. This can be used to log a command's stderr into the logger itself.
    """

    @abstractmethod
    def cmd(self, msg, cmd_name: str | None = None, *args, **kwargs) -> None:
        """
        Log a commands' captured output (maybe stderr or stdout)

        :param msg: The captured output.
        :param cmd_name: Which command name to register the command level to. If ``None`` then the default level-name
            is picked-up.
        """
        ...  # pragma: no cover


class WarningLogProtocol(Protocol):
    """
    Protocol supporting the warning method.
    """

    @abstractmethod
    def warning(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class ErrorLogProtocol(Protocol):
    """
    Protocol supporting the error method.
    """

    @abstractmethod
    def error(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class ExceptionLogProtocol(Protocol):
    """
    Protocol supporting the exception method.
    """

    @abstractmethod
    def exception(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class CriticalLogProtocol(Protocol):
    """
    Protocol supporting the critical method.
    """

    @abstractmethod
    def critical(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class FatalLogProtocol(Protocol):
    """
    Protocol supporting the critical method.
    """

    @abstractmethod
    def fatal(self, msg, *args, **kwargs) -> None: ...  # pragma: no cover


class _MinLogProtocol[L](
    LogLogProtocol[L],
    DebugLogProtocol,
    InfoLogProtocol,
    WarningLogProtocol,
    ErrorLogProtocol,
    CriticalLogProtocol,
    Protocol,
):
    """
    This logger protocol is designed for extension but not direct implementation.

    Useful when ``is-a`` relationship cannot be established between the interfaces that have most of the methods of
    each-other but conceptually do not behave in an ``is-a`` relationship.

    L - Level type, for e.g. ``int`` for python std logging.

    e.g.::

        AllLogProtocol has all the methods of MinLogProtocol but conceptually AllLogProtocol cannot be put in place
        of MinLogProtocol, i.e. there is no is-a relationship between them.


    Logger that has all the basic logging levels common to most (nearly all) loggers, i.e.::

        - DEBUG
        - INFO
        - WARNING
        - ERROR
        - CRITICAL
    """

    pass


class MinLogProtocol[L](_MinLogProtocol[L], Protocol):
    """
    Logger protocol that has all the basic logging levels common to most (nearly all) loggers, i.e.::

        - DEBUG
        - INFO
        - WARNING
        - ERROR
        - CRITICAL

    L - Level type, for e.g. ``int`` for python std logging.
    """

    pass


class AllLogProtocol[L](
    TraceLogProtocol,
    _MinLogProtocol[L],
    SuccessLogProtocol,
    NoticeLogProtocol,
    CommandLogProtocol,
    FatalLogProtocol,
    ExceptionLogProtocol,
    Protocol,
):
    """
    Logger protocol which supports all the common Logging levels, i.e.::

        - DEBUG
        - INFO
        - WARNING
        - ERROR
        - CRITICAL

    It also tries to add more levels that may facilitate users. Additionally supported log levels are::

        - TRACE
        - SUCCESS
        - NOTICE
        - COMMAND
        - FATAL
        - EXCEPTION

    L - Level type, for e.g. ``int`` for python std logging.
    """

    pass


class HasUnderlyingLogger[L](Protocol):
    """
    Insists that an underlying logger is contained in the class implementing this interface.

    Can return the contained underlying logger for the client class to perform actions in the future if needed.

    L - Level type, for e.g. ``int`` for python std logging.
    """

    @property
    @abstractmethod
    def underlying_logger(self) -> MinLogProtocol[L]:
        """
        It may not be a good idea to directly call this method to obtain underlying logger after class is
        initialised and its use is started. That is the case because that obtained underlying logger may tie the
        interfaces with a particular implementation and this will hinder in swapping logger implementations.

        :return: the contained underlying logger.
        """
        pass  # pragma: no cover


class SupportsTraceback(Protocol):
    """
    Can process (in most cases, log) the exception tracebacks.
    """

    @property
    @abstractmethod
    def traceback_enabled(self) -> bool:
        """
        :return: whether the traceback processing (in most cases, logging) is enabled.
        """
        ...


class AllLevelLogger[L](
    AllLogProtocol[L], HasUnderlyingLogger[L], SupportsTraceback, Protocol
):
    """
    Logger which supports all the common Logging levels, i.e.::

        - DEBUG
        - INFO
        - WARNING
        - ERROR
        - CRITICAL

    It also tries to add more levels that may facilitate users. Additionally supported log levels are::

        - TRACE
        - SUCCESS
        - NOTICE
        - COMMAND
        - FATAL
        - EXCEPTION

    And delegates the actual logging to an ``underlying_logger``, see ``HasUnderlyingLogger``.

    L - Level type, for e.g. ``int`` for python std logging.
    """

    pass



#!/usr/bin/env python3

"""
Configurators for python std logger handlers.
"""

import logging
from abc import abstractmethod
from typing import Protocol, IO, override

from logician.formatters import LogLevelFmt
from logician.stdlog.utils import form_stream_handlers_map, add_new_formatter
from logician.stdlog.constants import LOG_LVL as L, LOG_FMT as F


class HandlerConfigurator(Protocol):
    """
    Configure python stdlog's handlers.
    """

    @abstractmethod
    def configure(
        self,
        level: L,
        logger: logging.Logger,
        stream_fmt_map: dict[IO, LogLevelFmt[L, F]],
    ) -> None:
        """
        Logger handler's configurator.

        Configure the logger's handlers from a user supplied stream-fmt-map.

        :param level: L logging level.
        :param logger: the logger to configure.
        :param stream_fmt_map: the stream-format-handler-map that will configure the supplied logger's handlers.
        """
        pass  # pragma: no cover


class SimpleHandlerConfigurator(HandlerConfigurator):
    """
    Handler configurator that configures loggers' handlers using following rules:

    * Only have ``NullHandler`` in the logger if the supplied ``stream_fmt_map`` is empty. Supplying an empty
      ``stream_fmt_map`` shows the intent of the user to not log to any stream.
    * Add a ``StreamHandler`` for the stream that is provided in ``stream_fmt_map`` but is not present in the logger's
      handlers.
    * If handlers are already present and configured for a stream in logger then simply configure the first handler to
      conform to the supplied stream in the ``stream_fmt_map``.
    """

    @override
    def configure(
        self, level: L, logger: logging.Logger, stream_fmt_map: dict[IO, LogLevelFmt]
    ) -> None:
        """
        Logger handler's configurator.

        Configure the logger's handlers from a user supplied stream-fmt-map.

        >>> import sys
        >>> from logician.stdlog.formatters import StdLogAllLevelSameFmt

          * This configurator removes all the handlers if the supplied ``stream_fmt_map`` is empty {} or Falsy:

            * ``NullHandler`` remains to ensure that no logging is performed. User can specify their intent to not log by
              supplying ``stream_fmt_map`` as empty. This does not ``disable`` the logger, per se, just does not let it
              log to a stream.

            >>> lgr1 = logging.getLogger("lgr1")
            >>> SimpleHandlerConfigurator().configure(logging.INFO, lgr1,
            ...     {})      # empty stream_fmt_map to remove any handlers from lgr
            >>> assert isinstance(lgr1.handlers[0], logging.NullHandler)   # only NullHandler remains

          * Stream handler is added if handlers are not already present for a stream:

            >>> lgr2 = logging.getLogger("lgr2") # no handlers present for STDOUT stream by default
            >>> SimpleHandlerConfigurator().configure(logging.DEBUG, lgr2, {sys.stdout: StdLogAllLevelSameFmt()})  # configure and add handler for STDOUT stream
            >>> str_hn_map = form_stream_handlers_map(lgr2) # for the new stream->list[handlers] map
            >>> assert sys.stdout in str_hn_map # handler introduced for the STDOUT stream

          * Updates the first handler of a stream when the stream is supplied in the ``stream_fmt_map`` and logger already
            has a configured handler for that stream:

            >>> lgr2_1 = logging.getLogger("lgr2") # obtain lgr2 instance as it already has one STDOUT stream->handler configured from before
            >>> new_stdout_handler = logging.StreamHandler(sys.stdout)
            >>> lgr2_1.addHandler(new_stdout_handler) # add another handler for STDOUT stream
            >>> SimpleHandlerConfigurator().configure(logging.DEBUG, lgr2_1, {sys.stdout: StdLogAllLevelSameFmt("%(name)s")})
            >>> str_hn_map = form_stream_handlers_map(lgr2_1) # obtain the updated stream->list[handlers] map
            >>> assert "%(name)s" == str_hn_map[sys.stdout][0].formatter._fmt # type: ignore[attr-defined] only the 0th handler is configured

        :param level: L logging level.
        :param logger: the logger to configure.
        :param stream_fmt_map: the stream-format-handler-map that will configure the supplied logger's handlers.
        """

        if not stream_fmt_map:
            # empty user-supplied stream->formatter map
            # specifies the user's intent to not log anywhere hence, clear all existing handlers
            logger.handlers.clear()
            # add a NullHandler else the logging goes to logging.lastResort
            logger.addHandler(logging.NullHandler())
        else:
            stream_handlers_map = form_stream_handlers_map(logger)
            for stream in stream_fmt_map:
                lvl_fmt_handlr = stream_fmt_map[stream]
                fmt = lvl_fmt_handlr.fmt(level)  # obtain format for the required level
                if (
                    stream in stream_handlers_map
                ):  # handler already present for the current stream, as stream -> list[handler]
                    if stream_handlers_map[
                        stream
                    ]:  # handlers are already configured for this stream, as stream -> [handler1, handler2, ..., handlerN]
                        handlr = stream_handlers_map[stream][0]  # get the first handler
                        if handlr.formatter:  # handler already has a formatter
                            handlr.formatter._fmt = fmt
                        else:  # no formatter configured for the handler
                            handlr.setFormatter(
                                logging.Formatter(fmt=fmt)
                            )  # configure formatter for this handler
                    else:  # no handler configured for the required stream, as stream -> [], empty handler list or no handlers
                        logger.addHandler(add_new_formatter(stream, fmt))
                else:  # handlers not present for the current stream, as no stream->list[handlers] mapping present in the logger
                    # introduce a new handler
                    logger.addHandler(add_new_formatter(stream, fmt))


# TODO: hava a handler configurator that does not affect existing handlers of the logger which are configured outside
#  of logician



#!/usr/bin/env python3
# coding=utf-8


"""
Configurators for verbosity (V) and quietness (Q).
"""

# region configurators.vq.constants re-exports
from logician.configurators.vq.constants import Q_LITERAL as Q_LITERAL
from logician.configurators.vq.constants import V_LITERAL as V_LITERAL
from logician.configurators.vq.constants import VQ_DICT_LITERAL as VQ_DICT_LITERAL
# endregion

# region configurators.vq.base re-exports
from logician.configurators.vq.base import VQConfigurator as VQConfigurator
from logician.configurators.vq.base import VQLevelOrDefault as VQLevelOrDefault
# endregion

from logician.configurators.vq.sep import VQSepConfigurator as VQSepConfigurator

from logician.configurators.vq.comm import VQCommConfigurator as VQCommConfigurator



#!/usr/bin/env python3
# coding=utf-8

"""
A repo implementation for observability of logger configurators.
"""

import abc
import configparser
import json
import os
import tempfile
import tomllib
from abc import abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Protocol, Any, override

from logician.constants import LGCN_INFO_FP_ENV_VAR


class Persister[DS](Protocol):
    """
    A persister that can persist in-mem data of the ``DS`` data-structure format.

    ``DS`` - data-structure that this persister works with.
    """

    @abstractmethod
    def init(self):
        """
        Initialise the persistence storage.
        """

    @abstractmethod
    def commit(self, ds: DS):
        """
        Write ``ds`` to the persistent storage.

        :param ds: data-structure carrying writable payload.
        """
        ...

    @abstractmethod
    def reload(self) -> DS:
        """
        Read from persistent storage.

        :return: the read payload.
        """


class PathProvider(Protocol):
    @abstractmethod
    def get_path(self) -> Path: ...


class HasPathProvider(Protocol):
    @property
    @abstractmethod
    def path_provider(self) -> PathProvider: ...


class HandlerPathProvider(PathProvider, HasPathProvider, Protocol):
    pass


class FilePathProvider(HandlerPathProvider):
    def __init__(self, path_provider: "FilePathProvider"):
        self._path_provider = path_provider

    @abstractmethod
    def _get_file_path(self) -> Path | None: ...

    @override
    @property
    def path_provider(self) -> "FilePathProvider":
        return self._path_provider

    @override
    def get_path(self) -> Path:
        ret_path = self._get_file_path()
        return ret_path if ret_path else self.path_provider.get_path()


class EnvFilePathProvider(FilePathProvider):
    def __init__(self, env: str, path_provider: FilePathProvider):
        self.env = env
        super().__init__(path_provider)

    @override
    def _get_file_path(self) -> Path | None:
        env_val: str | None = os.getenv(self.env)
        return Path(env_val) if env_val is not None else None


class IniFilePathProvider(FilePathProvider):
    def __init__(
        self, path_provider: FilePathProvider, ini_file_path: Path = Path.cwd()
    ):
        super().__init__(path_provider)
        self.ini_file_path = Path(ini_file_path, "ittusa.ini")
        self.parser = configparser.ConfigParser()

    @override
    def _get_file_path(self) -> Path | None:
        if not self.ini_file_path.exists():
            return None
        self.parser.read_file(str(self.ini_file_path))
        file_path = self.parser.get("ittusa", "index_store")
        return Path(file_path) if file_path is not None else None


class PyprojectFilePathProvider(FilePathProvider):
    def __init__(
        self,
        path_provider: FilePathProvider,
        pyproject_root_file_path: Path = Path.cwd(),
    ):
        super().__init__(path_provider)
        self.pyproject_file_path = Path(pyproject_root_file_path, "pyproject.toml")

    @override
    def _get_file_path(self) -> Path | None:
        if not self.pyproject_file_path.exists():
            return None
        try:
            file_path = tomllib.loads(self.pyproject_file_path.read_text())["tool"][
                "ittusa"
            ]["ini"]["file_path"]
        except KeyError:
            return None
        return Path(file_path) if file_path is not None else None


class ConstTmpDirFPP(FilePathProvider):
    def __init__(
        self, file_path: Path = Path(tempfile.gettempdir(), ".0-LGCN-LOG-DETAILS.json")
    ):
        # Type ignoring arg of super().__init__() as FilePathProvider is required but None is provided
        super().__init__(None)  # type: ignore[arg-type]
        self.file_path = file_path

    @override
    def get_path(self) -> Path:
        return self._get_file_path()

    @override
    def _get_file_path(self) -> Path:
        return self.file_path


class FilePersister[DS](Persister, abc.ABC):
    def __init__(self, path_provider: PathProvider):
        self.file_path: Path = path_provider.get_path()

    def init(self):
        if not self.file_path.exists():
            self.file_path.write_text("")  # create the file


class JSONFilePersister(FilePersister[dict]):
    def commit(self, ds: dict):
        with open(self.file_path, mode="w+") as fp:
            json.dump(ds, fp)

    def reload(self) -> dict:
        try:
            return json.loads(self.file_path.read_text())
        except json.decoder.JSONDecodeError:
            return {}


class Repo(Protocol):
    """
    Repository that can:

    - initialise idempotently.
    - index (or store in memory) properties related to an object id.
    - commit the index to desired location. This is determined by the persister implementation.
    - reload (or refresh) the memory from the persister.
    - read the index (or memory) for retrieving properties relating to an object id.
    """

    @abstractmethod
    def init(self):
        """
        Initialise the repo.

        This will be an idempotent operation.

        It is required to be so because multiple classes/modules/packages may initialise the repo and it must not
        reinitialise its state every time.
        """
        ...

    @abstractmethod
    def index(self, id_: str, **attrs):
        """
        Store/Index the properties relating to ``id_`` in memory.

        :param id_: id to store properties for, in the memory.
        :param attrs: these properties will be stored against the ``id_``.
        """
        ...

    @abstractmethod
    def read(self, id_: str) -> dict[str, Any]:
        """
        Retrieve stored properties related to object ``_id`` from the index (or memory) without making expensive
        ``reload`` calls.

        Is an idempotent operation and does not contact the persister.

        :param id_: object id for which properties are to be queried from memory.
        :return: property-name -> property-value dictionary for the properties stored in-memory for object id ``id_``.
        """
        ...

    @abstractmethod
    def read_all(self) -> dict[str, dict[str, Any]]:
        """
        Retrieve all the stored properties related to all object ``_id`` from the index (or memory) without making
        expensive ``reload`` calls.

        Is an idempotent operation and does not contact the persister.

        :return: object-id -> {property-name -> property-value} dictionary for the properties stored in-memory for
            all object id(s).
        """
        ...

    @abstractmethod
    def reload(self):
        """
        Refresh the contents of memory or index by contacting the persister.
        """
        ...

    @abstractmethod
    def commit(self):
        """
        Persist in-memory to a persistent storage.

        Calls persister.
        """
        ...


class DictRepo(Repo):
    def __init__(self, persister: Persister[dict]):
        """
        Repo implementation using a ``defaultdict``.
        """
        self.repo: dict[str, dict[str, Any]] = defaultdict(dict)
        self.persister = persister

    @override
    def init(self):
        if self.repo is None:
            self.repo = defaultdict(dict)
        self.persister.init()

    @override
    def index(self, id_: str, **attrs):
        self.repo[id_].update(attrs)

    @override
    def read(self, id_: str) -> dict[str, Any]:
        return self.repo[id_].copy()

    def read_all(self) -> dict[str, dict[str, Any]]:
        return self.repo.copy()

    @override
    def commit(self):
        self.persister.commit(self.repo)

    @override
    def reload(self):
        self.repo = self.persister.reload()


__the_instance: Repo = DictRepo(
    JSONFilePersister(
        EnvFilePathProvider(
            LGCN_INFO_FP_ENV_VAR,
            EnvFilePathProvider(
                "ITTU_FP",
                IniFilePathProvider(PyprojectFilePathProvider(ConstTmpDirFPP())),
            ),
        )
    )
)


def get_repo() -> Repo:
    """
    :return: a singleton repo implementation.
    """
    return __the_instance



#!/usr/bin/env python3
# coding=utf-8

"""
Utility functions for logging related library.
"""


def command_or_file(command_name: str, file_name: str) -> str:
    """
    Useful in instances where a file may be either run as a library or a command. In such a case, mostly loggers must
    have appropriate names.

    Run as a library: A logger may be named as the file name (e.g. ``vt.utils.error.err_handler.err_funcs``)
    when it is run as a library or is used as a call from another file or command.

    Run as a command: When a file is run as a command then we'd need the logger to be named as the name of the command.
    Since files assume the name ``__main__`` when they're run directly as a command hence,the logger is then to be
    named as the command name.

    Examples:

    >>> command_or_file('a-cmd', 'vt.utils.error.err_handler.err_funcs')
    'vt.utils.error.err_handler.err_funcs'

    >>> command_or_file('a-cmd', '__main__')
    'a-cmd'

    :param command_name:
    :param file_name:
    :return: ``file_name`` if ``file_name`` is not ``__main__`` else ``command_name``.
    """
    return command_name if file_name == "__main__" else file_name



#!/usr/bin/env python3
# coding=utf-8


"""
Configure loggers as per level supplied by a list.

First non-None level in this supplied list is picked up by default.
"""

import logging
from typing import override

from logician import DirectStdAllLevelLogger
from logician._repo import get_repo
from logician.configurators import (
    LoggerConfigurator,
    HasUnderlyingConfigurator,
    LevelLoggerConfigurator,
)
from vt.utils.commons.commons.collections import get_first_non_none


class ListLoggerConfigurator[T](LoggerConfigurator, HasUnderlyingConfigurator):
    DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE = get_first_non_none

    def __init__(
        self,
        level_list: list[T | None],
        configurator: LevelLoggerConfigurator[T],
        level_pickup_strategy=DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE,
    ):
        """
        Picks up the first non ``None`` level from the supplied ``level_list`` to configure the logger underneath.

        * Examples:

          * supplied ``level_list`` cannot be ``None``:

            >>> ListLoggerConfigurator(None, # noqa: as level_list is deliberately passed as None
            ...     None) # noqa: as configurator is unused and passed as None
            Traceback (most recent call last):
            ValueError: Level list must not be None.

        :param level_list: list of log levels which may contain ``None``. First non-``None`` value
            is picked-up by default for logger configuration.
        :param configurator: configurator which is decorated by this logger-configurator.
        :param level_pickup_strategy: pick up a level from the list of levels supplied in ``level_list``. Default is
            to pick up the first non-``None`` level. ``DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE``.
        """
        if level_list is None:
            raise ValueError("Level list must not be None.")
        self._level_list = level_list
        self._underlying_configurator = configurator
        self.level_pickup_strategy = level_pickup_strategy
        get_repo().init()

    def configure(self, logger: logging.Logger) -> DirectStdAllLevelLogger:
        final_level = self.level_pickup_strategy(
            self.level_list, self.underlying_configurator.level
        )
        self.underlying_configurator.set_level(final_level)
        get_repo().index(logger.name, level_list=self.level_list, level=final_level)
        return self.underlying_configurator.configure(logger)

    @override
    @property
    def underlying_configurator(self) -> LevelLoggerConfigurator[T]:
        return self._underlying_configurator

    @property
    def level_list(self) -> list[T | None]:
        return self._level_list

    @override
    def clone(self, **overrides) -> "ListLoggerConfigurator[T]":
        """
        overrides:
            ``level_list`` - list of log levels which may contain ``None``. First non-``None`` value is picked-up by
            default for logger configuration.

            ``configurator`` - configurator which is decorated by this logger-configurator.

            ``level_pickup_strategy`` - pick up a level from the list of levels supplied in ``level_list``.
            Default is to pick up the first non-``None`` level. ``DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE``.
        :return: a new ``ListLoggerConfigurator``.
        """
        level_list = overrides.pop("level_list", self.level_list.copy())
        configurator = overrides.pop("configurator", self.underlying_configurator)
        level_pickup_strategy = overrides.pop(
            "level_pickup_strategy", self.level_pickup_strategy
        )
        return ListLoggerConfigurator[T](
            level_list, configurator, level_pickup_strategy
        )



#!/usr/bin/env python3
# coding=utf-8


"""
Logger interfaces for Logger configurators.
"""

# region configurators.base re-exports
from logician.configurators.base import LoggerConfigurator as LoggerConfigurator
from logician.configurators.base import (
    HasUnderlyingConfigurator as HasUnderlyingConfigurator,
)
from logician.configurators.base import (
    LevelLoggerConfigurator as LevelLoggerConfigurator,
)
from logician.configurators.base import LevelTarget as LevelTarget
# endregion



#!/usr/bin/env python3
# coding=utf-8


"""
Configurators for verbosity (V) and quietness (Q) that accept verbosity and quietness as one inclusive argument.

Useful when users know that verbosity and quietness will arrive in one parameter, for e.g. a CLI which accepts
environment variables as such:

- LGCN_ALL_LOG=vv
- CCC_LIB_LOG = qq

Here the value for a simple environment variable maybe vv or qq and hence can arrive in a single parameter.
"""

from abc import abstractmethod
from typing import Protocol, override

from logician.configurators.vq import (
    VQConfigurator,
    V_LITERAL,
    Q_LITERAL,
    VQ_DICT_LITERAL,
    VQLevelOrDefault,
)
from logician.configurators.vq.base import SimpleWarningVQLevelOrDefault


class VQCommConfigurator[T](VQConfigurator[T], Protocol):
    """
    Configurator which takes verbosity and quietness commonly (in a single argument) for configuration.
    """

    @abstractmethod
    def validate(self, ver_qui: V_LITERAL | Q_LITERAL | None) -> bool:
        """
        Validate whether the supplied verbosity or quietness are valid.

        :param ver_qui: verbosity or quietness.
        :return: ``True`` if inputs are valid, ``False`` otherwise.
        :raise ValueError: if values for ``ver_qui`` are invalid and subclass decides to raise the error.
        """
        ...  # pragma: no cover

    @abstractmethod
    def get_effective_level(
        self, ver_qui: V_LITERAL | Q_LITERAL | None, default_level: T
    ) -> T:
        """
        Get the effective level for supplied verbosity or quietness.

        :param ver_qui: verbosity or quietness.
        :param default_level: returned if both verbosity or quietness are ``None`` or not supplied.
        :return: computed level for verbosity and quietness or ``default_level`` if verbosity or quietness
            are ``None``.
        :raise KeyError: if the verbosity or quietness is not found in the ``vq_level_map`` and the subclass decides
            to raise error for this.
        """
        ...  # pragma: no cover


class VQCommon[T](VQCommConfigurator[T]):
    def __init__(
        self,
        vq_level_map: VQ_DICT_LITERAL[T],
        warn_only: bool = False,
        level_or_default_handler: VQLevelOrDefault[T] | None = None,
    ):
        """
        Treats verbosity and quietness as one inclusive argument.

        Treats such conditions as an error::

            - supplied verbosity or quietness is not within the ``vq_level_map``.

        :param vq_level_map: A dictionary containing verbosity|quietness -> logging.level mapping.
        :param warn_only: Only warn on potential errors instead of raising an Error.
        :param level_or_default_handler: Level computer. Defaults to ``SimpleWarningVQLevelOrDefault`` if ``None`` or
            not supplied.
        """
        self._vq_level_map = vq_level_map
        if level_or_default_handler:
            self.level_or_default_handler = level_or_default_handler
        else:
            self.level_or_default_handler = SimpleWarningVQLevelOrDefault(
                vq_level_map, warn_only=warn_only
            )

    @override
    @property
    def vq_level_map(self) -> VQ_DICT_LITERAL[T]:
        return self._vq_level_map

    @override
    def validate(self, ver_qui: V_LITERAL | Q_LITERAL | None) -> bool:
        """
        Examples::

        >>> VQCommon({}).validate('v')
        False


        >>> VQCommon({}).validate('q')
        False

        >>> VQCommon({}).validate('vqn') # noqa
        False

        >>> VQCommon[int]({'v': 20}).validate('v')
        True


        >>> VQCommon({'v': 20, 'q': 1}).validate('q')
        True

        :return: ``True`` if ``ver_qui`` is in the assigned level map, else ``False``.
        """
        if ver_qui in self.vq_level_map:
            return True
        return False

    @override
    def get_effective_level(
        self, ver_qui: V_LITERAL | Q_LITERAL | None, default_level: T
    ) -> T:
        """
        Get effective level by treating verbosity and quietness as single argument.

        Examples::

        >>> import sys
        >>> import contextlib
        >>> import warnings

        Level inquiry::

            Get queried verbosity:

            >>> VQCommon[int]({'v': 20}).get_effective_level('v', 10)
            20

            Return default_level if queried verbosity is not registered and warn_only is True:

            >>> with warnings.catch_warnings(record=True) as w:
            ...     VQCommon[int]({'v': 20, 'q': 1}, True).get_effective_level('vv', 10)
            ...     assert "'vv': Unexpected verbosity or quietness value. Choose from 'v' and 'q'." in str(w.pop().message)
            10

            Raise KeyError if queried verbosity or quietness is not registered and warn_only is False or not provided:

            >>> VQCommon[int]({'v': 20, 'qqq': 70}).get_effective_level('qq', 10)
            Traceback (most recent call last):
            ...
            KeyError: "'qq': Unexpected verbosity or quietness value. Choose from 'v' and 'qqq'."

        :param ver_qui: `verbosity` or `quietness`.
        :param default_level: level to return when verbosity and quietness are not in the ``vq_level_map``.
        :returns: corresponding logging level according to verbosity and quietness calculations.
            ``default_level`` if both verbosity and quietness are ``None`` or not supplied.
            ``default_level`` if both are not supplied and ``warn_only`` is ``True``.
        :rtype T: a type
        :raise KeyError: if verbosity and quietness are absent in ``vq_level_map`` and ``warn_only`` is ``False``.
        """
        if ver_qui:
            level = self.level_or_default_handler.level_or_default(
                ver_qui,
                "verbosity or quietness",
                default_level,
                list(self.vq_level_map.keys()),
            )
        else:
            level = default_level
        return level



#!/usr/bin/env python3
# coding=utf-8

"""
Classes w.r.t implementation inheritance are defined here.
"""

import warnings
from abc import abstractmethod
from logging import Logger
from typing import override, Protocol

from logician.delegating import AllLevelLoggerImplABC
from logician.stdlog import (
    TRACE_LOG_LEVEL,
    NOTICE_LOG_LEVEL,
    SUCCESS_LOG_LEVEL,
    StdLogProtocol,
    INDIRECTION_STACK_LEVEL,
    FATAL_LOG_LEVEL,
    CMD_LOG_LEVEL,
    CMD_LOG_STR,
    EXCEPTION_TRACEBACK_LOG_LEVEL,
)
from logician.stdlog.utils import TempSetLevelName
from logician.stdlog.constants import LOG_LVL as L


class StdProtocolAllLevelLoggerImpl(AllLevelLoggerImplABC[L], Protocol):
    """
    Interface for all logging levels provided by the standard logging protocol.
    """

    @override
    @property
    @abstractmethod
    def underlying_logger(self) -> StdLogProtocol:
        pass

    @override
    @abstractmethod
    def cmd(self, msg, cmd_name: str | None = None, *args, **kwargs) -> None:
        """
        Log a commands' captured output (maybe stderr or stdout)

        :param msg: The captured output.
        :param cmd_name: Which command name to register the command level to. If ``None`` then the default level-name
            ``COMMAND`` is picked-up.
        """
        ...


class BaseDirectStdAllLevelLoggerImpl(StdProtocolAllLevelLoggerImpl, Protocol):
    """
    Interface for all logging levels provided by the python standard logging library.
    """

    @override
    @property
    @abstractmethod
    def underlying_logger(self) -> Logger:  # noqa
        pass


class DirectAllLevelLoggerImpl(BaseDirectStdAllLevelLoggerImpl):
    def __init__(self, underlying_logger: Logger, stack_level=INDIRECTION_STACK_LEVEL):
        """
        Basic logger that implements all the logging levels of python standard logging and simply delegates method
        calls to the underlying logger. Created for implementation inheritance.

        :param underlying_logger: logger (python standard logger) that actually performs the logging.
        :param stack_level: stack to go up to get the file/line/func information from the framing stack.
            Check ``DEFAULT_STACK_LEVEL`` for more details.
        """
        self._underlying_logger = underlying_logger
        self.stack_level = stack_level

    @override
    @property
    def underlying_logger(self) -> Logger:  # noqa
        return self._underlying_logger

    @override
    @property
    def traceback_enabled(self) -> bool:
        """
        >>> import logging

        Examples:

          * Traceback enables when log level is equal to ``EXCEPTION_TRACEBACK_LOG_LEVEL``:

            >>> lgr = logging.getLogger("tb-demo-1")
            >>> lgr.setLevel(EXCEPTION_TRACEBACK_LOG_LEVEL)
            >>> logger = DirectAllLevelLoggerImpl(lgr)
            >>> assert logger.traceback_enabled

          * Traceback enables when log level is less than ``EXCEPTION_TRACEBACK_LOG_LEVEL``:

            >>> lgr = logging.getLogger("tb-demo-2")
            >>> lgr.setLevel(EXCEPTION_TRACEBACK_LOG_LEVEL-1)
            >>> logger = DirectAllLevelLoggerImpl(lgr)
            >>> assert logger.traceback_enabled

          * Traceback disables when log level is more than ``EXCEPTION_TRACEBACK_LOG_LEVEL``:

            >>> lgr = logging.getLogger("tb-demo-2")
            >>> lgr.setLevel(EXCEPTION_TRACEBACK_LOG_LEVEL+10)
            >>> logger = DirectAllLevelLoggerImpl(lgr)
            >>> assert not logger.traceback_enabled

        :return: whether the traceback processing (in most cases, logging) is enabled.
        """
        return self.underlying_logger.level <= EXCEPTION_TRACEBACK_LOG_LEVEL

    @override
    def trace(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.log(
            TRACE_LOG_LEVEL, msg, *args, stacklevel=self.stack_level, **kwargs
        )

    @override
    def debug(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.debug(msg, *args, stacklevel=self.stack_level, **kwargs)

    @override
    def info(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.info(msg, *args, stacklevel=self.stack_level, **kwargs)

    @override
    def success(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.log(
            SUCCESS_LOG_LEVEL, msg, *args, stacklevel=self.stack_level, **kwargs
        )

    @override
    def notice(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.log(
            NOTICE_LOG_LEVEL, msg, *args, stacklevel=self.stack_level, **kwargs
        )

    @override
    def cmd(self, msg, cmd_name: str | None = None, *args, **kwargs) -> None:
        if self.underlying_logger.isEnabledFor(CMD_LOG_LEVEL):
            with TempSetCmdLvlName(cmd_name):
                self.underlying_logger.log(
                    CMD_LOG_LEVEL, msg, *args, stacklevel=self.stack_level, **kwargs
                )

    @override
    def warning(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.warning(
            msg, *args, stacklevel=self.stack_level, **kwargs
        )

    @override
    def error(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.error(msg, *args, stacklevel=self.stack_level, **kwargs)

    @override
    def critical(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.critical(
            msg, *args, stacklevel=self.stack_level, **kwargs
        )

    @override
    def fatal(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.log(
            FATAL_LOG_LEVEL, msg, *args, stacklevel=self.stack_level, **kwargs
        )

    @override
    def exception(self, msg, *args, **kwargs) -> None:
        self.underlying_logger.exception(
            msg, *args, exc_info=True, stacklevel=self.stack_level, **kwargs
        )

    @override
    def log(self, level: L, msg: str, *args, **kwargs) -> None:
        self.underlying_logger.log(
            level, msg, *args, stacklevel=self.stack_level, **kwargs
        )


class TempSetCmdLvlName(TempSetLevelName):
    def __init__(self, cmd_name: str | None, no_warn: bool = False):
        """
        Set the command log level name temporarily and then revert it back to the ``CMD_LOG_STR``.

        :param cmd_name: Command log Level name to set the level to.
        :param no_warn: A warning is shown if the supplied ``level_name`` is strip-empty. This warning can be suppressed
            by setting ``no_warn=True``.
        """
        super().__init__(CMD_LOG_LEVEL, cmd_name, CMD_LOG_STR, no_warn)

    @override
    def _warn_user(self):
        warnings.warn(
            f"Supplied log level name for command log level {self.level} is empty."
        )



#!/usr/bin/env python3
# coding=utf-8


"""
Logger configurators that configure log levels using environment variables.
"""

import logging
import os
from typing import override, cast

from logician import DirectStdAllLevelLogger
from logician._repo import get_repo
from logician.constants import LGCN_ALL_LOG_ENV_VAR
from logician.configurators import LevelLoggerConfigurator
from logician.configurators.list_lc import ListLoggerConfigurator


class EnvListLC[T](ListLoggerConfigurator[T]):
    DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE = ListLoggerConfigurator[
        T
    ].DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE

    def __init__(
        self,
        env_list: list[str],
        configurator: LevelLoggerConfigurator[T],
        level_pickup_strategy=DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE,
    ):
        """
        Environment variable list logger configurator.

        This logger configurator can be used to configure log level using values supplied from environment variables.
        Default behavior is to pick up the first passed environment variable value. Designed to process log level from
        multiple environment variables and hence has a precedence order to the values form environment variables. The
        first environment variable value takes highest precedence and then the precedence diminishes.

        :param env_list: list of environment variables. Default behavior is to take precedence in decreasing order.
        :param configurator: underlying logger configurator.
        :param level_pickup_strategy: strategy to pick-up level from a supplied list of levels. Default is to pick up
            the first supplied, then next and then so on.
        """
        super().__init__([], configurator, level_pickup_strategy)
        self._env_list = env_list
        get_repo().init()

    @property
    def env_list(self) -> list[str]:
        return self._env_list

    @override
    @property
    def level_list(self) -> list[T | None]:
        return [cast(T | None, os.getenv(e)) for e in self.env_list]

    @override
    def configure(self, logger: logging.Logger) -> DirectStdAllLevelLogger:
        get_repo().index(logger.name, env_list=self.env_list)
        return super().configure(logger)

    @override
    def clone(self, **overrides) -> "EnvListLC[T]":
        """
        overrides:
            ``env_list`` - list of environment variables. Default behavior is to take precedence in decreasing order.

            ``configurator`` - configurator which is decorated by this logger-configurator.

            ``level_pickup_strategy`` - pick up a level from the list of levels supplied in ``level_list``. Default is
            to pick up the first non-``None`` level. ``DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE``.
        :return: a new ``EnvListLC``.
        """
        env_list = overrides.pop("env_list", self.env_list.copy())
        configurator = overrides.pop("configurator", self.underlying_configurator)
        level_pickup_strategy = overrides.pop(
            "level_pickup_strategy", self.level_pickup_strategy
        )
        return EnvListLC[T](env_list, configurator, level_pickup_strategy)

    def clone_with_envs(
        self, env: str, *envs: str, low_precedence: bool = False
    ) -> "EnvListLC[T]":
        """
        Clone the current environment list level logger configurator with some extra environment variables. May be used
        in scenarios when a certain module needs a Logger configurator dependent on the environment variables of another
        logger configurator and wants to include its own environment variable as well in the mix.

        For e.g. the `push-pull-prep` project has its logger configurator to heed to ``ENV_PPP`` environment variable
        and `push-pull-prep.some_other_module.py` needs to support ``ENV_PPP.SOM`` environment variable along with the
        parent module (`push-pull-prep`'s) environment variable (``ENV_PPP``). Then it can do so like this:

        Examples:

        * `push-pull-prep`'s environment variable logger configurator:

        >>> ppp_lc = EnvListLC(['ENV_PPP'],
        ...     None) # noqa: as configurator is unused and passed as None
        >>> assert ['ENV_PPP'] == ppp_lc.env_list

        * `push-pull-prep.some_other_module.py`'s environment variable logger configurator, which builds upon the
        `push-pull-prep`'s environment variable logger configurator. It can add its own environment variable
        ``ENV_PPP.SOM`` and by default that takes the highest precedence:

        >>> som_lc = ppp_lc.clone_with_envs('ENV_PPP.SOM')
        >>> assert ['ENV_PPP.SOM', 'ENV_PPP'] == som_lc.env_list # some_lc retains the original ppp_lc's env-var (ENV_PPP) along with its own, but its own env-var (ENV_PPP.SOM) has a higher precedence
        >>> assert ['ENV_PPP'] == ppp_lc.env_list # no change to the original logger configurator's env list.

        * Add multiple env vars:

        >>> som_lc = ppp_lc.clone_with_envs('SUMO', 'ENV_PPP.SOM', 'ENV_PPP.SOM.MOS') # multiple env vars can be registered.
        >>> assert ['SUMO', 'ENV_PPP.SOM', 'ENV_PPP.SOM.MOS', 'ENV_PPP'] == som_lc.env_list
        >>> assert ['ENV_PPP'] == ppp_lc.env_list # no change to the original logger configurator's env list.

        * Add vars with lower precedence than the env vars of the original or parent logger configurator by setting
        ``low_precedence=True``:

        >>> som_lc = ppp_lc.clone_with_envs('OTHER_ENV', low_precedence=True)
        >>> assert ['ENV_PPP', 'OTHER_ENV'] == som_lc.env_list
        >>> assert ['ENV_PPP'] == ppp_lc.env_list # no change to the original logger configurator's env list

        >>> som_lc = ppp_lc.clone_with_envs('SUMO', 'ENV_PPP.SOM', low_precedence=True)
        >>> assert ['ENV_PPP', 'SUMO', 'ENV_PPP.SOM'] == som_lc.env_list
        >>> assert ['ENV_PPP'] == ppp_lc.env_list # no change to the original logger configurator's env list.

        :param env: extra environment variables which need to be introduced over and above the original logger
            configurator's own env vars.
        :param envs: extra environment variables which need to be introduced over and above the original logger
            configurator's own env vars.
        :param low_precedence: the extra introduced envs mostly prepend the original logger configurator's env vars and
            thus have a higher precedence than them. Setting this param as ``True`` makes the extra introduced env vars
            append (not prepend) the original logger configurator's env vars and thus keeps them at a lower precedence
            than the original logger configurator's env vars.
        :return: new logger configurator with extra newly introduced env vars.
        """
        if low_precedence:
            _env_list = [*self.env_list, env, *envs]
        else:
            _env_list = [env, *envs, *self.env_list]
        return self.clone(env_list=_env_list)


class LgcnEnvListLC[T](EnvListLC[T]):
    DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE = EnvListLC[
        T
    ].DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE

    def __init__(
        self,
        env_list: list[str],
        configurator: LevelLoggerConfigurator[T],
        level_pickup_strategy=DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE,
        all_log_env_var: str = LGCN_ALL_LOG_ENV_VAR,
    ):
        """
        LgcnEnvListLC -> Logician Env var List Logger Configurator.

        This logger configurator can be used to configure log level using values supplied from environment variables.
        Default behavior is to pick up the first passed environment variable value. Designed to process log level from
        multiple environment variable values and hence has a precedence order to the values from environment variables.
        The first environment variable value takes highest precedence and then the precedence diminishes. Environment
        variable ``LGCN_ALL_LOG`` is always appended to ``env_list`` so that if no environment variable is registered
        then at least this one is registered.

        :param env_list: list of environment variables. Default behavior is to take precedence in decreasing order.
        :param configurator: underlying logger configurator.
        :param level_pickup_strategy: strategy to pick-up level from a supplied list of levels. Default is to pick up
            the first supplied, then next and then so on.
        :param all_log_env_var: Environment variable which, by default, will be checked last to get the logging levels.
        """
        env_list.append(all_log_env_var)
        super().__init__(env_list, configurator, level_pickup_strategy)

    @override
    def clone(self, **overrides) -> "LgcnEnvListLC[T]":
        """
        overrides:
            ``env_list`` - list of environment variables. Default behavior is to take precedence in decreasing order.

            ``configurator`` - configurator which is decorated by this logger-configurator.

            ``level_pickup_strategy`` - pick up a level from the list of levels supplied in ``level_list``. Default is
            to pick up the first non-``None`` level. ``DEFAULT_LEVEL_PICKUP_FIRST_NON_NONE``.
        :return: a new ``LgcnEnvListLC``.
        """
        level_list = overrides.pop("env_list", self.env_list.copy())
        configurator = overrides.pop("configurator", self.underlying_configurator)
        level_pickup_strategy = overrides.pop(
            "level_pickup_strategy", self.level_pickup_strategy
        )
        return LgcnEnvListLC[T](level_list, configurator, level_pickup_strategy)



#!/usr/bin/env python3
# coding=utf-8

"""
All logging interface implementation by the standard logging library of python.

All level of logging are supported by these loggers using delegation and indirection
"""

from abc import ABC, abstractmethod
from logging import Logger
from typing import override, cast, Protocol

from logician import AllLevelLogger
from logician.delegating import DelegatingLogger
from logician.stdlog import StdLogProtocol
from logician.stdlog.base import DirectStdAllLevelLogger
from logician.stdlog.all_levels_impl import (
    StdProtocolAllLevelLoggerImpl,
    BaseDirectStdAllLevelLoggerImpl,
)
from logician.stdlog.constants import LOG_LVL as L, LOG_STR_LVL as S


class StdProtocolAllLevelLogger(AllLevelLogger[L], DelegatingLogger[L], Protocol):
    """
    Interface for a std protocol logger which provides all logging levels by the protocol implementation.
    """

    @override
    @property
    @abstractmethod
    def logger_impl(self) -> StdProtocolAllLevelLoggerImpl: ...

    @override
    @property
    @abstractmethod
    def underlying_logger(self) -> StdLogProtocol: ...


class BaseStdProtocolAllLevelLogger(StdProtocolAllLevelLogger, ABC):
    def __init__(self, logger_impl: StdProtocolAllLevelLoggerImpl):
        """
        Implementation for a std protocol logger which provides all logging levels by the protocol implementation.

        see ``StdProtocolAllLevelLoggerImpl``.

        :param logger_impl: the logger implementations where all logging calls will be forwarded to.
        """
        self._logger_impl = logger_impl
        self._underlying_logger = self._logger_impl.underlying_logger
        self.name = self._logger_impl.underlying_logger.name
        self.level = self._logger_impl.underlying_logger.level
        self.disabled = self._logger_impl.underlying_logger.disabled

    @override
    @property
    def logger_impl(self) -> StdProtocolAllLevelLoggerImpl:
        return self._logger_impl

    @override
    @property
    def underlying_logger(self) -> StdLogProtocol:
        return self._underlying_logger  # pragma: no cover

    @override
    @property
    def traceback_enabled(self) -> bool:
        return self.logger_impl.traceback_enabled  # pragma: no cover

    @override
    def trace(self, msg, *args, **kwargs) -> None:
        self.logger_impl.trace(msg, *args, **kwargs)

    @override
    def debug(self, msg, *args, **kwargs) -> None:
        self.logger_impl.debug(msg, *args, **kwargs)

    @override
    def info(self, msg, *args, **kwargs) -> None:
        self.logger_impl.info(msg, *args, **kwargs)

    @override
    def notice(self, msg, *args, **kwargs) -> None:
        self.logger_impl.notice(msg, *args, **kwargs)

    @override
    def success(self, msg, *args, **kwargs) -> None:
        self.logger_impl.success(msg, *args, **kwargs)

    @override
    def cmd(self, msg, cmd_name: str | None = None, *args, **kwargs) -> None:
        """
        Log a commands' captured output (maybe stderr or stdout)

        :param msg: The captured output.
        :param cmd_name: Which command name to register the command level to. If ``None`` then the default level-name
            ``COMMAND`` is picked-up. But as this is a ``DelegatingLogger`` hence this behavior can be altered in the
            delegatee class.
        """
        self.logger_impl.cmd(msg, cmd_name, *args, **kwargs)

    @override
    def warning(self, msg, *args, **kwargs) -> None:
        self.logger_impl.warning(msg, *args, **kwargs)

    @override
    def error(self, msg, *args, **kwargs) -> None:
        self.logger_impl.error(msg, *args, **kwargs)

    @override
    def critical(self, msg, *args, **kwargs) -> None:
        self.logger_impl.critical(msg, *args, **kwargs)

    @override
    def fatal(self, msg, *args, **kwargs) -> None:
        self.logger_impl.fatal(msg, *args, **kwargs)

    @override
    def exception(self, msg, *args, **kwargs) -> None:
        self.logger_impl.exception(msg, *args, **kwargs)

    @override
    def log(self, level: L, msg: str, *args, **kwargs) -> None:
        self.logger_impl.log(level, msg, *args, **kwargs)


class BaseDirectStdAllLevelLogger(
    BaseStdProtocolAllLevelLogger, DirectStdAllLevelLogger, ABC
):
    def __init__(
        self,
        logger_impl: BaseDirectStdAllLevelLoggerImpl,
        level_name_map: dict[L, S] | None = None,
        cmd_name: str | None = None,
    ):
        """
        Implementation for a std protocol logger which provides all logging levels by the protocol implementation.

        see ``BaseDirectStdAllLevelLoggerImpl``.

        This class also registers the log-level->log-level-name if one is supplied. But doesn't do any
        level registration if log-level->log-level-name is empty or not supplied.

        ``DirectStdAllLevelLogger.register_levels()`` can be called to initialise and register default levels prior
        to creating this class's object.

        :param logger_impl: the logger implementations where all logging calls will be forwarded to.
        :param level_name_map: a log-level->log-level-name map. eg: ``30`` -> ``INFO``. This is useful for registering
            the level->name map with the logger.
        :param cmd_name: level name for the command-log-level. ``None`` specifies that the default ``COMMAND`` will
            be shown on log.cmd() call. But as this is a ``DelegatingLogger`` hence this behavior can be altered in the
            delegatee class.
        """
        super().__init__(logger_impl)
        if level_name_map:
            BaseDirectStdAllLevelLogger.register_levels(level_name_map)
        self.level_name_map = level_name_map
        self.cmd_name = cmd_name

    @override
    @property
    def logger_impl(self) -> BaseDirectStdAllLevelLoggerImpl:
        return cast(BaseDirectStdAllLevelLoggerImpl, self._logger_impl)

    @override
    @property
    def underlying_logger(self) -> Logger:  # noqa
        return cast(Logger, self._underlying_logger)

    @override
    def cmd(self, msg, cmd_name: str | None = None, *args, **kwargs) -> None:
        """
        Log a commands' captured output (maybe stderr or stdout)

        :param msg: The captured output.
        :param cmd_name: Which command name to register the command logging level to. If ``None`` then the ctor-supplied
            level-name ``self.cmd_name`` is picked-up. If ``self.cmd_name`` is also ``None`` then default
            ``COMMAND`` is picked-up. But as this is a ``DelegatingLogger`` hence this behavior can be altered in the
            delegatee class.
        """
        final_cmd_name = cmd_name or self.cmd_name
        self.logger_impl.cmd(msg, final_cmd_name, *args, **kwargs)


class DirectAllLevelLogger(BaseDirectStdAllLevelLogger):
    def __init__(
        self,
        logger_impl: BaseDirectStdAllLevelLoggerImpl,
        level_name_map: dict[L, S] | None = None,
        cmd_name: str | None = None,
    ):
        """
        Std protocol logger which provides all logging levels by the protocol implementation.

        see ``BaseDirectStdAllLevelLoggerImpl``.

        This class also registers the log-level->log-level-name if one is supplied. But doesn't do any
        level registration if log-level->log-level-name is empty or not supplied.

        ``DirectStdAllLevelLogger.register_levels()`` can be called to initialise and register default levels prior
        to creating this class's object.

        :param logger_impl: the logger implementations where all logging calls will be forwarded to.
        :param level_name_map: a log-level->log-level-name map. eg: ``30`` -> ``INFO``. This is useful for registering
            the level->name map with the logger.
        :param cmd_name: level name for the command-log-level. ``None`` specifies that the default ``COMMAND`` will
            be shown on log.cmd() call. But as this is a ``DelegatingLogger`` hence this behavior can be altered in the
            delegatee class.
        """
        super().__init__(logger_impl, level_name_map, cmd_name)

    @override
    @property
    def logger_impl(self) -> BaseDirectStdAllLevelLoggerImpl:
        return cast(BaseDirectStdAllLevelLoggerImpl, self._logger_impl)



#!/usr/bin/env python3
# coding=utf-8


"""
Logger interfaces for standard Logger formatters.
"""

import logging
import sys
from typing import override, IO, Protocol

from logician.formatters import AllLevelSameFmt, DiffLevelDiffFmt, LogLevelFmt
from logician.stdlog import (
    TIMED_DETAIL_LOG_FMT,
    TRACE_LOG_LEVEL,
    DETAIL_LOG_FMT,
    SHORT_LOG_FMT,
    SHORTER_LOG_FMT,
    SUCCESS_LOG_LEVEL,
)
from logician.stdlog.constants import LOG_LVL as L, LOG_FMT as F


class StdLogLevelFmt(LogLevelFmt[L, F], Protocol):
    """
    Base interface for all the level-format mappers for stdlog.
    """

    pass


class StdLogAllLevelSameFmt(StdLogLevelFmt, AllLevelSameFmt[L, F]):
    def __init__(self, fmt: F = SHORTER_LOG_FMT):
        """
        Same std log format for all levels.

        :param fmt: logging format constant for all std log levels.
        """
        self._fmt = fmt

    @override
    def fmt(self, level: L) -> F:
        return self._fmt


class StdLogAllLevelDiffFmt(StdLogLevelFmt, DiffLevelDiffFmt[L, F]):
    DEFAULT_LOGGER_DICT: dict[L, F] = {
        TRACE_LOG_LEVEL: TIMED_DETAIL_LOG_FMT,
        logging.DEBUG: DETAIL_LOG_FMT,
        logging.INFO: SHORT_LOG_FMT,
        SUCCESS_LOG_LEVEL: SHORTER_LOG_FMT,
    }
    """
    Different log formats for different log levels.
    """

    def __init__(self, fmt_dict: dict[L, F] | None = None):
        """
        Specify how different log levels should impact the logging formats.

        For e.g.::

            - least verbose ERROR level.
            ERROR: an error occurred.

            - less verbose INFO level.
            logger.name: INFO: some information

            - verbose DEBUG level.
            logger.name: DEBUG: [filename.py - func()]: some debug info

            - most verbose TRACE level.
            2025-04-03 20:59:39,418: TRACE: [filename.py:218 - func()]: some trace info

        provides immediately-upper registered level if an unregistered level is queried.

        :param fmt_dict: level -> format dictionary. Defaults to
            ``StdLogAllLevelDiffFmt.DEFAULT_LOGGER_DICT`` when ``None`` or an empty dict is provided.
        """
        self._fmt_dict = (
            fmt_dict if fmt_dict else StdLogAllLevelDiffFmt.DEFAULT_LOGGER_DICT
        )

    @override
    def fmt(self, level: L) -> F:
        final_level = (
            level if level in self._fmt_dict else self.next_approx_level(level)
        )
        return self._fmt_dict[final_level]

    @override
    def next_approx_level(self, missing_level: L) -> L:
        """
        :param missing_level: A level that was not registered in the logger.
        :return: immediately-upper registered level if a ``missing_level`` is queried.
        """
        max_level = max(self._fmt_dict)
        if missing_level >= max_level:
            return max_level

        for level in sorted(self._fmt_dict.keys()):
            if level > missing_level:
                return level
        return max_level


def stderr_all_lvl_same_fmt(fmt: F | None = None) -> dict[IO, StdLogLevelFmt]:
    """
    Examples:

      * ``SHORTER_LOG_FMT`` is used when the ``fmt`` param is not supplied:

        >>> fmt_dict = stderr_all_lvl_same_fmt()
        >>> assert sys.stderr in fmt_dict
        >>> assert fmt_dict[sys.stderr].fmt(logging.DEBUG) == SHORTER_LOG_FMT

      * Supplied format is used when ``fmt`` is supplied:

        >>> fmt_dict = stderr_all_lvl_same_fmt("%(name)s")
        >>> assert sys.stderr in fmt_dict
        >>> assert fmt_dict[sys.stderr].fmt(logging.FATAL) == "%(name)s"

    :param fmt: the format required for all levels
    :return: stderr->same-format-for-all-levels dict.
    """
    if fmt is not None:
        return {sys.stderr: StdLogAllLevelSameFmt(fmt)}
    return {sys.stderr: StdLogAllLevelSameFmt()}


STDERR_ALL_LVL_SAME_FMT: dict[IO, StdLogLevelFmt] = stderr_all_lvl_same_fmt()
"""
Maps ``sys.stderr`` to same logging format for all levels.
"""

STDERR_ALL_LVL_DIFF_FMT: dict[IO, StdLogLevelFmt] = {
    sys.stderr: StdLogAllLevelDiffFmt()
}
"""
Maps ``sys.stderr`` to different logging format for all levels.
"""



from vt.utils.errors.error_specs.exceptions import (
    VTException,
    VTExitingException,
    VTCmdException,
    VTCmdNotFoundError,
)


class LogicianException(VTException):
    """
    Exception particular to ``logician``.
    """

    pass


class LogicianExitingException(LogicianException, VTExitingException):
    """
    A ``logician`` exception that allows exiting with an ``error_code``.
    """

    pass


class LogicianCmdException(LogicianExitingException, VTCmdException):
    """
    A ``logician`` exception that can denote exceptional scenario from a command run.
    """

    pass


class LogicianCmdNotFoundError(LogicianExitingException, VTCmdNotFoundError):
    """
    A ``logician`` exception that specifies a command not found.
    """

    pass



#!/usr/bin/env python3
# coding=utf-8


"""
Global constants for the logician.
"""

LGCN_ALL_LOG_ENV_VAR = "LGCN_ALL_LOG"
"""
LGCN_ALL_LOG_ENV_VAR -> Logician All Log Env Var.

Set logging level for all logging that applies to Logician.
"""

LGCN_MAIN_CMD_NAME = "lgcn"
"""
CLI command name for logician CLI.
"""

LGCN_INFO_FP_ENV_VAR = "LGCN_INFO_FP"
"""
Environment variable which will be used by LGCN CLI to talk to other programs to gather their logger configurators'
details.
"""



#!/usr/bin/env python3
# coding=utf-8


"""
Configurators for verbosity (V) and quietness (Q) that accept verbosity and quietness as exclusive separate arguments.

Useful when users know that verbosity and quietness will arrive in separate parameters, for e.g. a CLI which accepts
-v, -vv ... etc. for verbosity and -qq, -qqq ... etc. for quietness.
"""

from abc import abstractmethod
from typing import Protocol, override

from vt.utils.errors.warnings import vt_warn

from logician import errmsg_creator
from logician.configurators.vq import (
    VQConfigurator,
    V_LITERAL,
    Q_LITERAL,
    VQ_DICT_LITERAL,
    VQLevelOrDefault,
)
from logician.configurators.vq.base import SimpleWarningVQLevelOrDefault


class VQSepConfigurator[T](VQConfigurator[T], Protocol):
    """
    Configurator which takes verbosity and quietness separately (as separate arguments) for configuration.
    """

    @abstractmethod
    def validate(
        self, verbosity: V_LITERAL | None, quietness: Q_LITERAL | None
    ) -> bool:
        """
        Validate whether the supplied verbosity and quietness are valid.

        :param verbosity: verbosity.
        :param quietness: quietness.
        :return: ``True`` if inputs are valid, ``False`` otherwise.
        """
        ...  # pragma: no cover

    @abstractmethod
    def get_effective_level(
        self, verbosity: V_LITERAL | None, quietness: Q_LITERAL | None, default_level: T
    ) -> T:
        """
        Get the effective level for supplied verbosity and quietness.

        :param verbosity: verbosity.
        :param quietness: quietness.
        :param default_level: returned if both verbosity and quietness are ``None`` or not supplied.
        :return: computed level for verbosity and quietness or ``default_level`` if both verbosity and quietness
            are ``None``.
        """
        ...  # pragma: no cover


class VQSepExclusive[T](VQSepConfigurator[T]):
    def __init__(
        self,
        vq_level_map: VQ_DICT_LITERAL[T],
        warn_only: bool = False,
        level_or_default_handler: VQLevelOrDefault[T] | None = None,
    ):
        """
        Treats verbosity and quietness as separate and exclusive, i.e. both cannot be given together.

        Treats such conditions as an error::

            - verbosity and quietness are provided together.
            - supplied verbosity or quietness is not within the ``vq_level_map``.

        :param vq_level_map: A dictionary containing verbosity|quietness -> logging.level mapping.
        :param warn_only: Only warn on potential errors instead of raising an Error.
        """
        self._vq_level_map = vq_level_map
        self.warn_only = warn_only
        if level_or_default_handler:
            self.level_or_default_handler = level_or_default_handler
        else:
            self.level_or_default_handler = SimpleWarningVQLevelOrDefault(
                vq_level_map, warn_only=warn_only
            )

    @override
    @property
    def vq_level_map(self) -> VQ_DICT_LITERAL[T]:
        return self._vq_level_map

    @override
    def validate(
        self, verbosity: V_LITERAL | None, quietness: Q_LITERAL | None
    ) -> bool:
        """
        ``verbosity`` and ``quietness`` are not accepted together.

        Examples::

        >>> import sys
        >>> import contextlib
        >>> import warnings

        Raise error if ``warn_only`` is ``False`` or not provided::

        >>> VQSepExclusive({}).validate('v', 'q')
        Traceback (most recent call last):
        ...
        ValueError: verbosity and quietness are not allowed together

        Only warn if ``warn_only`` is provided ``True``::

        >>> with warnings.catch_warnings(record=True) as w:
        ...     VQSepExclusive({}, True).validate('v', 'q')
        ...     assert "verbosity and quietness are not allowed together" in str(w.pop().message)
        False

        Return ``True`` if only one verbosity is supplied::

        >>> VQSepExclusive({}).validate('v', None)
        True

        Return ``True`` if only one quietness is supplied::

        >>> VQSepExclusive({}).validate(None, 'q')
        True

        Return ``True`` if both are ``None``::

        >>> VQSepExclusive({}).validate(None, None)
        True

        :raise ValueError: if both verbosity and quietness are given and If ``self.warn_only`` is ``False``.
        :return: If ``self.warn_only`` is ``True`` - ``True`` if inputs are valid, ``False`` otherwise.
        """
        if verbosity and quietness:
            warn_str = errmsg_creator.not_allowed_together("verbosity", "quietness")
            if self.warn_only:
                vt_warn(warn_str)
                return False
            else:
                raise ValueError(warn_str)
        else:
            return True

    @override
    def get_effective_level(
        self, verbosity: V_LITERAL | None, quietness: Q_LITERAL | None, default_level: T
    ) -> T:
        """
        Get effective level by treating verbosity and quietness as separate entities.

        Note::

            verbosity and quietness are not to be provided together.

        Examples::

        >>> import sys
        >>> import contextlib
        >>> import warnings

        Both verbosity and quietness provided together::

            warn_only is not provided or is False:

            >>> VQSepExclusive({}).get_effective_level('v', 'q', 10)
            Traceback (most recent call last):
            ...
            ValueError: verbosity and quietness are not allowed together

            Only warn if warn_only is provided True and return the default_value:

            >>> with warnings.catch_warnings(record=True) as w:
            ...     VQSepExclusive[int]({'v': 20}, True).get_effective_level('v', 'q', 10)
            ...     assert "verbosity and quietness are not allowed together" in str(w.pop().message)
            10

        Level inquiry::

            Get queried verbosity:

            >>> VQSepExclusive[int]({'v': 20}).get_effective_level('v', None, 10)
            20

            Return default_level if queried verbosity is not registered and warn_only is True:

            >>> with warnings.catch_warnings(record=True) as w:
            ...     VQSepExclusive[int]({'v': 20}, True).get_effective_level('vv', None, 10)
            ...     assert "'vv': Unexpected verbosity value. Choose from 'v'." in str(w.pop().message)
            10

            Raise KeyError if queried verbosity is not registered and warn_only is False or not provided:

            >>> VQSepExclusive[int]({'v': 20}).get_effective_level('vv', None, 10)
            Traceback (most recent call last):
            ...
            KeyError: "'vv': Unexpected verbosity value. Choose from 'v'."

        :param verbosity: verbosity.
        :param quietness: quietness.
        :param default_level: level to return when verbosity and quietness are not in the ``vq_level_map``.
        :returns: corresponding logging level according to verbosity and quietness calculations.
            ``default_level`` if both verbosity and quietness are ``None`` or not supplied.
            ``default_level`` if both are not supplied and ``warn_only`` is ``True``.
        :rtype T: a type
        :raise KeyError: if verbosity and quietness are absent in ``vq_level_map`` and ``warn_only`` is ``False``.
        :raise ValueError: If both verbosity and quietness are given and ``self.warn_only`` is ``False``.
        """
        if not self.validate(verbosity, quietness):
            level = default_level
        else:
            if verbosity:
                level = self.level_or_default_handler.level_or_default(
                    verbosity,
                    "verbosity",
                    default_level,
                    list(self.vq_level_map.keys()),
                )
            elif quietness:
                level = self.level_or_default_handler.level_or_default(
                    quietness,
                    "quietness",
                    default_level,
                    list(self.vq_level_map.keys()),
                )
            else:
                level = default_level
        return level



#!/usr/bin/env python3
# coding=utf-8


"""
Configure loggers as per level supplied by a supplier.
"""

import logging
from typing import Callable, override

from logician import DirectStdAllLevelLogger
from logician._repo import get_repo
from logician.configurators import (
    LoggerConfigurator,
    HasUnderlyingConfigurator,
    LevelLoggerConfigurator,
)


class SupplierLoggerConfigurator[T](LoggerConfigurator, HasUnderlyingConfigurator):
    def __init__(
        self,
        level_supplier: Callable[[], T | None],
        configurator: LevelLoggerConfigurator[T],
    ):
        """
        Configurator that configures loggers as per the level supplied by the ``level_supplier``.

        Examples:

        >>> from logician.stdlog.configurator import StdLoggerConfigurator

          * Typed:

            >>> _ = SupplierLoggerConfigurator[int](lambda: logging.DEBUG, StdLoggerConfigurator())

          * Untyped

            >>> _ = SupplierLoggerConfigurator(lambda : logging.INFO, StdLoggerConfigurator())

        :param level_supplier: a supplier to supply level.
        :param configurator: underlying configurator.
        """
        self.level_supplier = level_supplier
        self._underlying_configurator = configurator
        get_repo().init()

    def configure(self, logger: logging.Logger) -> DirectStdAllLevelLogger:
        """
        >>> from logician.stdlog.configurator import StdLoggerConfigurator
        >>> from logician.stdlog import NOTICE_LOG_LEVEL

        Examples:

          * Level set by the supplier: ``INFO`` level

            >>> _lgr = logging.getLogger("supplier-logger-demo-1")
            >>> lc = SupplierLoggerConfigurator(lambda : logging.INFO, StdLoggerConfigurator())
            >>> lgr = lc.configure(_lgr)
            >>> assert lgr.underlying_logger.level == logging.INFO

          * Supplier returns ``None`` hence, level is determined by the underlying configurator: ``NOTICE`` level

            >>> _lgr = logging.getLogger("supplier-logger-demo-2")
            >>> lc = StdLoggerConfigurator(level=NOTICE_LOG_LEVEL)
            >>> assert lc.level == NOTICE_LOG_LEVEL
            >>> lc = SupplierLoggerConfigurator(lambda : None, lc)
            >>> lgr = lc.configure(_lgr)
            >>> assert lgr.underlying_logger.level == NOTICE_LOG_LEVEL

        :param logger: the logger to configure.
        :return: configured logger with its logging level set by the ``level_supplier``.
        """
        computed_level = self.level_supplier()
        final_level = (
            computed_level
            if computed_level is not None
            else self.underlying_configurator.level
        )
        self.underlying_configurator.set_level(final_level)
        get_repo().index(logger.name, level=final_level)
        return self.underlying_configurator.configure(logger)

    @override
    @property
    def underlying_configurator(self) -> LevelLoggerConfigurator[T]:
        return self._underlying_configurator  # pragma: no cover

    @override
    def clone(self, **overrides) -> "SupplierLoggerConfigurator":
        """
        overrides:
            ``level_supplier`` - a supplier to supply level.

            ``configurator`` - underlying configurator.

        Examples:

        >>> import logging
        >>> from logician.stdlog.configurator import StdLoggerConfigurator

          * Simple clone, no overrides:

            >>> lc1 = SupplierLoggerConfigurator(lambda : logging.INFO, StdLoggerConfigurator())
            >>> lc2 = lc1.clone()
            >>> assert lc1.underlying_configurator == lc2.underlying_configurator and lc1.level_supplier == lc2.level_supplier

          * Clone, override just ``level_supplier``:

            >>> lc1 = SupplierLoggerConfigurator(lambda : logging.INFO, StdLoggerConfigurator())
            >>> lc2 = lc1.clone(level_supplier=lambda : logging.DEBUG)
            >>> assert lc1.underlying_configurator == lc2.underlying_configurator and lc1.level_supplier != lc2.level_supplier

          * Clone, override just ``configurator``:

            >>> lc1 = SupplierLoggerConfigurator(lambda : logging.INFO, StdLoggerConfigurator())
            >>> lc2 = lc1.clone(configurator=StdLoggerConfigurator(level=logging.CRITICAL))
            >>> assert lc1.underlying_configurator != lc2.underlying_configurator and lc1.level_supplier == lc2.level_supplier

          * Clone, override all:

            >>> lc1 = SupplierLoggerConfigurator(lambda : None, StdLoggerConfigurator())
            >>> lc2 = lc1.clone(level_supplier=lambda : logging.WARNING, configurator=StdLoggerConfigurator(level=logging.CRITICAL))
            >>> assert lc1.underlying_configurator != lc2.underlying_configurator and lc1.level_supplier != lc2.level_supplier

        :return: a new ``SupplierLoggerConfigurator``.
        """
        level_supplier = overrides.pop("level_supplier", self.level_supplier)
        configurator = overrides.pop("configurator", self.underlying_configurator)
        return SupplierLoggerConfigurator(level_supplier, configurator)



#!/usr/bin/env python3
# coding=utf-8

"""
Constants related to logging implementation by the python standard logging.
"""

import logging

LOG_LVL = int
"""
Python stdlog level type.

E.g. 10 or logging.DEBUG
"""

LOG_STR_LVL = str
"""
Python stdlog str level type.

E.g. "WARNING"
"""

type EX_LOG_LVL = LOG_LVL | LOG_STR_LVL
"""
Extended log level type.

Extended Type Var for python standard logger logging level.

E.g. logging.DEBUG->"DEBUG"
"""

LOG_FMT = str
"""
Python stdlog logger format type.

E.g. "%(name)s -- %(message)s"
"""

DEFAULT_STACK_LEVEL = 2
"""
``2`` chosen as value because the classes that use this constant actually delegate logging to a user supplied 
underlying_logger. The underlying_logger uses stacklevel=1 to get the 
immediate calling stack, i.e. function, source, filename, lineno, etc of the immediate caller of the log method as in
log.debug(). But when this is delegated by another class (log-capturing-class) to underlying std logger then the 
source information of the log-capturing-class is preferred as that is stacklevel=1, which we do not want. We want that
the source information of the caller-class must be shown. This is just one level up in the calling stack hence, 
stacklevel=DEFAULT_STACK_LEVEL=2 is chosen.


Illustration for ``stacklevel=1``
---------------------------------

    log_caller_src.py :: log.info('info msg') -> log_capturing_class.py :: self.underlying_logger.info('info msg')
    
Prints ::

    log_capturing_class.py | INFO | underlying_logger.info | info msg


Illustration for ``stacklevel=2``
---------------------------------

    log_caller_src.py :: log.info('info msg') -> log_capturing_class.py :: self.underlying_logger.info('info msg')
    
Prints ::

    log_caller_class.py | INFO | calling_meth | info msg

"""


INDIRECTION_STACK_LEVEL = DEFAULT_STACK_LEVEL + 1
"""
Just one more level up on the stack for checking logs.

see DEFAULT_STACK_LEVEL for more details on this.
"""

TRACE_LOG_LEVEL = (
    logging.DEBUG - 5
)  # 5, previous level at 3 -> TRACEBACK; next level at 10 -> DEBUG
TRACE_LOG_STR = "TRACE"

EXCEPTION_TRACEBACK_LOG_LEVEL = TRACE_LOG_LEVEL - 2  # 3, next level at 5 -> TRACEBACK
"""
Exception traces should only be printed when the user really wants to dig deep into the code and hence should have
very low log level
"""
EXCEPTION_TRACEBACK_LOG_STR = "TRACEBACK"

CMD_LOG_LEVEL = (
    logging.DEBUG + 5
)  # 15, previous level at 10 -> DEBUG; next level at 20 -> INFO
CMD_LOG_STR = "COMMAND"

SUCCESS_LOG_LEVEL = logging.INFO + 3  # 23
SUCCESS_LOG_STR = "SUCCESS"

NOTICE_LOG_LEVEL = SUCCESS_LOG_LEVEL + 3  # 26
NOTICE_LOG_STR = "NOTICE"

FATAL_LOG_LEVEL = logging.CRITICAL + 10
FATAL_LOG_STR = "FATAL"
SHORTER_LOG_FMT = "%(levelname)s: %(message)s"
SHORT_LOG_FMT = "%(name)s: %(levelname)s: %(message)s"
DETAIL_LOG_FMT = (
    "%(name)s: %(levelname)s: [%(filename)s - %(funcName)10s() ]: %(message)s"
)
TIMED_DETAIL_LOG_FMT = (
    "%(asctime)s: %(name)s: %(levelname)s: [%(filename)s:%(lineno)d - "
    "%(funcName)10s() ]: %(message)s"
)

WARNING_LEVEL: LOG_LVL = logging.WARNING
"""
Default logging level for the python std lib.
"""

DEFAULT_LOG_LEVEL_SUCCESS = SUCCESS_LOG_LEVEL
"""
Default log level for all std loggers.
"""



#!/usr/bin/env python3
# coding=utf-8

"""
Logging implemented by python standard logging.
"""

# region stdlog.constants re-exports
from logician.stdlog.constants import DEFAULT_STACK_LEVEL as DEFAULT_STACK_LEVEL
from logician.stdlog.constants import INDIRECTION_STACK_LEVEL as INDIRECTION_STACK_LEVEL
from logician.stdlog.constants import TRACE_LOG_LEVEL as TRACE_LOG_LEVEL
from logician.stdlog.constants import TRACE_LOG_STR as TRACE_LOG_STR
from logician.stdlog.constants import SUCCESS_LOG_LEVEL as SUCCESS_LOG_LEVEL
from logician.stdlog.constants import SUCCESS_LOG_STR as SUCCESS_LOG_STR
from logician.stdlog.constants import NOTICE_LOG_LEVEL as NOTICE_LOG_LEVEL
from logician.stdlog.constants import NOTICE_LOG_STR as NOTICE_LOG_STR
from logician.stdlog.constants import CMD_LOG_LEVEL as CMD_LOG_LEVEL
from logician.stdlog.constants import CMD_LOG_STR as CMD_LOG_STR
from logician.stdlog.constants import (
    EXCEPTION_TRACEBACK_LOG_LEVEL as EXCEPTION_TRACEBACK_LOG_LEVEL,
)
from logician.stdlog.constants import (
    EXCEPTION_TRACEBACK_LOG_STR as EXCEPTION_TRACEBACK_LOG_STR,
)
from logician.stdlog.constants import FATAL_LOG_LEVEL as FATAL_LOG_LEVEL
from logician.stdlog.constants import FATAL_LOG_STR as FATAL_LOG_STR
from logician.stdlog.constants import SHORTER_LOG_FMT as SHORTER_LOG_FMT
from logician.stdlog.constants import SHORT_LOG_FMT as SHORT_LOG_FMT
from logician.stdlog.constants import DETAIL_LOG_FMT as DETAIL_LOG_FMT
from logician.stdlog.constants import TIMED_DETAIL_LOG_FMT as TIMED_DETAIL_LOG_FMT
from logician.stdlog.constants import WARNING_LEVEL as WARNING_LEVEL
# endregion

# region stdlog.base re-exports
from logician.stdlog.base import StdLevelLogger as StdLevelLogger
from logician.stdlog.base import StdLogProtocol as StdLogProtocol
from logician.stdlog.base import DirectStdAllLevelLogger as DirectStdAllLevelLogger
# endregion

# region stdlog.all_levels re-exports
from logician.stdlog.all_levels import (
    StdProtocolAllLevelLogger as StdProtocolAllLevelLogger,
)
from logician.stdlog.all_levels import (
    BaseDirectStdAllLevelLogger as BaseDirectStdAllLevelLogger,
)
from logician.stdlog.all_levels import DirectAllLevelLogger as DirectAllLevelLogger
# endregion



#!/usr/bin/env python3
# coding=utf-8


"""
Base interfaces for verbosity (V) and quietness (Q) configurators.
"""

from abc import abstractmethod
from typing import Protocol, Literal, Any, override, overload

from vt.utils.errors.error_specs import DefaultOrError, WarningWithDefault
from vt.utils.errors.error_specs.base import SimpleWarningWithDefault
from vt.utils.errors.warnings import Warner

from logician.configurators.vq import VQ_DICT_LITERAL, V_LITERAL, Q_LITERAL

from logician import errmsg_creator


class VQConfigurator[T](Protocol):
    """
    Configurator for verbosity and quietness configurations.
    """

    @property
    @abstractmethod
    def vq_level_map(self) -> VQ_DICT_LITERAL[T]:
        """
        :return: A dictionary containing verbosity|quietness -> logging.level mapping.
        """
        ...  # pragma: no cover


class VQLevelOrDefault[T](VQConfigurator[T], Protocol):
    """
    Interface to facilitate getting a logging level from VQConfigurator.
    """

    def level_or_default(
        self,
        ver_qui: V_LITERAL | Q_LITERAL | None,
        emphasis: Literal["verbosity", "quietness", "verbosity or quietness"],
        default_level: T,
        choices: list[Any],
    ) -> T:
        """
        :param ver_qui: verbosity or quietness.
        :param emphasis: strings '`verbosity`' or '`quietness`'.
        :param default_level: logging level to be returned if ``ver_qui`` is ``None``.
        :param choices: What are the choices for `verbosity` or `quietness` or 'verbosity or quietness'.
        :return: calculated logging level from ``ver_qui`` or ``default_level`` if ``ver_qui`` is ``None``.
        :raise KeyError: if verbosity and quietness are absent in ``vq_level_map`` and ``self.key_error_handler``
            decides to re raise the error.
        """
        if ver_qui:
            try:
                return self.vq_level_map[ver_qui]
            except KeyError as e:
                return self.key_error_handler.handle_key_error(
                    e, default_level, emphasis, choices
                )
        else:
            return default_level

    @property
    @abstractmethod
    def key_error_handler(self) -> DefaultOrError[T]: ...


class SimpleWarningVQLevelOrDefault[T](VQLevelOrDefault[T], Warner):
    @overload
    def __init__(self, vq_level_map: VQ_DICT_LITERAL[T]): ...

    @overload
    def __init__(self, vq_level_map: VQ_DICT_LITERAL[T], *, warn_only: bool): ...

    @overload
    def __init__(
        self,
        vq_level_map: VQ_DICT_LITERAL[T],
        *,
        key_error_handler: WarningWithDefault[T],
    ): ...

    def __init__(
        self,
        vq_level_map: VQ_DICT_LITERAL[T],
        *,
        warn_only: bool | None = None,
        key_error_handler: WarningWithDefault[T] | None = None,
    ):
        """
        Simple implementation for ``VQLevelOrDefault``. It can decided by the ``key_error_handler`` on how to handle
        ``KeyError``.

        Default behavior is to simply warn the user on occurrence of a ``KeyError``.

        Fail-fast error examples::

        >>> from vt.utils.errors.error_specs.base import NoErrWarningWithDefault
        >>> vq_levels: VQ_DICT_LITERAL[int] = {'v': 10, 'vv': 5, 'q': 30}

        >>> SimpleWarningVQLevelOrDefault[int](vq_levels, warn_only=False, key_error_handler=NoErrWarningWithDefault())
        Traceback (most recent call last):
        ...
        ValueError: warn_only and key_error_handler are not allowed together


        Correct no error examples::

        >>> _ = SimpleWarningVQLevelOrDefault[int](vq_levels)

        >>> _ = SimpleWarningVQLevelOrDefault[int](vq_levels, warn_only=False)

        >>> _ = SimpleWarningVQLevelOrDefault[int](vq_levels, key_error_handler=NoErrWarningWithDefault())

        :param vq_level_map: verbosity-quietness level map.
        :param warn_only: warn the user of an error? Cannot be provided with ``key_error_handler``.
        :param key_error_handler: A custom ``KeyError`` to handle key errors.
        """
        if warn_only is not None and key_error_handler is not None:
            raise ValueError(
                errmsg_creator.not_allowed_together("warn_only", "key_error_handler")
            )
        self._vq_level_map = vq_level_map
        if key_error_handler:
            self._key_error_handler = key_error_handler
        elif warn_only is not None:
            self._key_error_handler = SimpleWarningWithDefault[T](warn_only=warn_only)
        else:
            # default behavior to just warn user on KeyError.
            self._key_error_handler = SimpleWarningWithDefault[T](warn_only=True)
        self._warn_only = self.key_error_handler.warn_only

    @override
    @property
    def vq_level_map(self) -> VQ_DICT_LITERAL[T]:
        return self._vq_level_map

    @override
    @property
    def warn_only(self) -> bool:
        return self._warn_only  # pragma: no cover

    @override
    @property
    def key_error_handler(self) -> WarningWithDefault[T]:
        return self._key_error_handler



#!/usr/bin/env python3
# coding=utf-8


"""
Mappers for

IO-streams -> log-level-format

refer ``logician.formatters.LogLevelFmt`` for log-level-format.
"""

from abc import abstractmethod
from typing import Protocol, IO

from logician.formatters import LogLevelFmt


class StreamFormatMapperComputer[L, F](Protocol):
    """
    Interface for the strategies that can compute and then generate mappings of stream -> level-format-mapper.

    level-format-map - see ``logician.formatters.LogLevelFmt``.

    L - type of ``level`` of the logger, e.g. ``int`` for python std log, like, 10 or logging.DEBUG

    F – type of the returned format, e.g. ``str`` for python std log, like "%(name)s - %(message)s"
    """

    @abstractmethod
    def compute(
        self, same_fmt_per_lvl: F | bool | None, stream_set: set[IO] | None
    ) -> dict[IO, LogLevelFmt[L, F]]:
        """
        Compute the stream format mapper from supplied arguments.

        :param same_fmt_per_lvl: Want same format per logging level?
        :param stream_set: Set of streams this format configuration is to be applied to. Note that ``{}`` denoting an
            empty stream_set is accepted and specifies the user's intent of not logging to any stream.
        :return: a dictionary of stream->log-level-format.
        """

        pass  # pragma: no cover



#!/usr/bin/env python3
# coding=utf-8

"""
Logging interfaces for the standard logging library of python.
"""

import logging
from abc import abstractmethod
from typing import Protocol, Any, Mapping, override

from logician import MinLogProtocol, AllLevelLogger
from logician.base import FatalLogProtocol, ExceptionLogProtocol, HasUnderlyingLogger
from logician.stdlog import (
    TRACE_LOG_LEVEL,
    TRACE_LOG_STR,
    SUCCESS_LOG_LEVEL,
    SUCCESS_LOG_STR,
    NOTICE_LOG_LEVEL,
    NOTICE_LOG_STR,
    EXCEPTION_TRACEBACK_LOG_LEVEL,
    EXCEPTION_TRACEBACK_LOG_STR,
    FATAL_LOG_LEVEL,
    FATAL_LOG_STR,
    CMD_LOG_LEVEL,
    CMD_LOG_STR,
)
from logician.stdlog.utils import level_name_mapping
from logician.stdlog.constants import LOG_LVL as L, LOG_STR_LVL as S


class StdLogProtocol(MinLogProtocol[L], Protocol):
    """
    Logger protocol that is followed (for methods) by the python std logging.

    Two additional methods are added on top of the MinLogProtocol::

        - fatal
        - exception

    along with properties that python std logger provides::

        - name
        - level
        - disabled
    """

    name: S
    level: L
    disabled: bool

    def fatal(self, msg: str, *args, **kwargs) -> None: ...

    # noinspection SpellCheckingInspection
    # required for the param stack-level because this method signature from the protocol needs to correctly match that
    # of the std logging method signature.
    def exception(
        self,
        msg: object,
        *args: object,
        exc_info: Any = ...,
        stack_info: bool = ...,
        stacklevel: L = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...


class StdLevelLogger(
    MinLogProtocol[L],
    FatalLogProtocol,
    ExceptionLogProtocol,
    HasUnderlyingLogger[L],
    Protocol,
):
    """
    Logger that implements python standard logging methods::

        - debug
        - info
        - warning
        - error
        - critical
        - fatal
        - exception
    """

    pass


class DirectStdAllLevelLogger(AllLevelLogger[L], Protocol):
    """
    All logging levels as provided by the python std logging.
    """

    DEFAULT_LEVEL_MAP: dict[L, S] = {
        TRACE_LOG_LEVEL: TRACE_LOG_STR,
        SUCCESS_LOG_LEVEL: SUCCESS_LOG_STR,
        NOTICE_LOG_LEVEL: NOTICE_LOG_STR,
        CMD_LOG_LEVEL: CMD_LOG_STR,
        EXCEPTION_TRACEBACK_LOG_LEVEL: EXCEPTION_TRACEBACK_LOG_STR,
        FATAL_LOG_LEVEL: FATAL_LOG_STR,
    }
    """
    All log levels in accordance with the python std log. Ordered in such a fashion::
    
        3 -> TRACEBACK
        5 -> TRACE
        10 -> DEBUG
        20 -> INFO
        23 -> SUCCESS
        26 -> NOTICE
        28 -> COMMAND
        30 -> WARNING
        40 -> ERROR
        50 -> CRITICAL
        60 -> FATAL
    """

    @staticmethod
    def register_levels(level_name_map: dict[L, S] | None = None) -> dict[L, S]:
        """
        Register levels in the python std logger.

        Note::

            The level changes are global in python std library hence, multiple calls to
            ``DirectStdAllLevelLogger.register_levels()`` may result in the latest call to win.

        * Examples:

          * Register more levels:

            >>> lvl_name_map = DirectStdAllLevelLogger.register_levels({1:"TRACE_DETAILED", 70: "BOMBED"})
            >>> assert lvl_name_map[1] == "TRACE_DETAILED"
            >>> assert lvl_name_map[70] == "BOMBED"

          * Change name for existing levels

            >>> lvl_name_map = level_name_mapping()
            >>> assert lvl_name_map[logging.DEBUG] == "DEBUG"
            >>> lvl_name_map = DirectStdAllLevelLogger.register_levels({logging.DEBUG: "DETAILED_INFO"})
            >>> assert lvl_name_map[logging.DEBUG] == "DETAILED_INFO" # DEBUG level name changed

          * Original (Default) levels kept if no levels are supplied:

            >>> lvl_name_map = DirectStdAllLevelLogger.register_levels()
            >>> assert all(lvl_name_map[lvl]==lvl_name for lvl, lvl_name in DirectStdAllLevelLogger.DEFAULT_LEVEL_MAP.items())

          * Original (Default) levels kept if empty level dict is provided:

            >>> lvl_name_map = DirectStdAllLevelLogger.register_levels()
            >>> assert all(lvl_name_map[lvl]==lvl_name for lvl, lvl_name in DirectStdAllLevelLogger.DEFAULT_LEVEL_MAP.items())

        :param level_name_map: log level - name mapping. This mapping updates the
            ``DirectStdAllLevelLogger.DEFAULT_LEVEL_MAP`` and then all the updated
            ``DirectStdAllLevelLogger.DEFAULT_LEVEL_MAP`` log levels are registered.
        :return: An ascending sorted level -> name map of all the registered log levels.
        """
        if level_name_map:
            DirectStdAllLevelLogger.DEFAULT_LEVEL_MAP.update(level_name_map)
        DirectStdAllLevelLogger.__register_all_levels(
            DirectStdAllLevelLogger.DEFAULT_LEVEL_MAP
        )
        return level_name_mapping()

    @staticmethod
    def __register_all_levels(level_name_map: dict[L, S]):
        for level in level_name_map:
            logging.addLevelName(level, level_name_map[level])

    @override
    @property
    @abstractmethod
    def underlying_logger(self) -> logging.Logger:  # noqa
        pass



#!/usr/bin/env python3
# coding=utf-8


"""
Logger interfaces for standard Logger configurators.
"""

import logging
from typing import override, overload, Protocol, IO

from vt.utils.errors.warnings import vt_warn

from logician import DirectAllLevelLogger, DirectStdAllLevelLogger
from logician import errmsg_creator
from logician._repo import get_repo
from logician.configurators import (
    LoggerConfigurator,
    HasUnderlyingConfigurator,
    LevelLoggerConfigurator,
)
from logician.configurators.vq import (
    V_LITERAL,
    Q_LITERAL,
    VQ_DICT_LITERAL,
    VQConfigurator,
    VQSepConfigurator,
    VQCommConfigurator,
)
from logician.configurators.vq.comm import VQCommon
from logician.configurators.vq.sep import VQSepExclusive
from logician.format_mappers import StreamFormatMapperComputer
from logician.formatters import LogLevelFmt
from logician.stdlog import TRACE_LOG_LEVEL, FATAL_LOG_LEVEL
from logician.stdlog.all_levels_impl import DirectAllLevelLoggerImpl
from logician.stdlog.format_mappers import StdStrFmtMprComputer
from logician.stdlog.hndlr_cfgr import HandlerConfigurator, SimpleHandlerConfigurator
from logician.stdlog.constants import (
    EX_LOG_LVL as E,
    LOG_LVL as L,
    LOG_FMT as F,
    LOG_STR_LVL as S,
    DEFAULT_LOG_LEVEL_SUCCESS,
)


class StdLoggerConfigurator(LevelLoggerConfigurator[E]):
    LOG_LEVEL_DEFAULT_SUCCESS = DEFAULT_LOG_LEVEL_SUCCESS
    CMD_NAME_NONE = None
    STREAM_FMT_MAPPER_NONE = None
    FMT_PER_LEVEL_NONE = None
    STREAM_SET_NONE = None
    LEVEL_NAME_MAP_NONE = None
    NO_WARN_FALSE = False
    PROPAGATE_FALSE = False

    @overload
    def __init__(
        self,
        *,
        level: E = LOG_LEVEL_DEFAULT_SUCCESS,
        cmd_name: str | None = CMD_NAME_NONE,
        same_fmt_per_lvl: F | bool | None = FMT_PER_LEVEL_NONE,
        stream_set: set[IO] | None = STREAM_SET_NONE,
        level_name_map: dict[L, S] | None = LEVEL_NAME_MAP_NONE,
        no_warn: bool = NO_WARN_FALSE,
        propagate: bool = PROPAGATE_FALSE,
        handlr_cfgr: HandlerConfigurator = SimpleHandlerConfigurator(),
        stream_fmt_mapper_computer: StreamFormatMapperComputer[
            L, F
        ] = StdStrFmtMprComputer(),
    ): ...

    @overload
    def __init__(
        self,
        *,
        level: E = LOG_LEVEL_DEFAULT_SUCCESS,
        cmd_name: str | None = CMD_NAME_NONE,
        stream_fmt_mapper: dict[IO, LogLevelFmt[L, F]] | None = STREAM_FMT_MAPPER_NONE,
        level_name_map: dict[L, S] | None = LEVEL_NAME_MAP_NONE,
        no_warn: bool = NO_WARN_FALSE,
        propagate: bool = PROPAGATE_FALSE,
        handlr_cfgr: HandlerConfigurator = SimpleHandlerConfigurator(),
        stream_fmt_mapper_computer: StreamFormatMapperComputer[
            L, F
        ] = StdStrFmtMprComputer(),
    ): ...

    def __init__(
        self,
        *,
        level: E = LOG_LEVEL_DEFAULT_SUCCESS,
        cmd_name: str | None = CMD_NAME_NONE,
        stream_fmt_mapper: dict[IO, LogLevelFmt[L, F]] | None = STREAM_FMT_MAPPER_NONE,
        same_fmt_per_lvl: F | bool | None = FMT_PER_LEVEL_NONE,
        stream_set: set[IO] | None = STREAM_SET_NONE,
        level_name_map: dict[L, S] | None = LEVEL_NAME_MAP_NONE,
        no_warn: bool = NO_WARN_FALSE,
        propagate: bool = PROPAGATE_FALSE,
        handlr_cfgr: HandlerConfigurator = SimpleHandlerConfigurator(),
        stream_fmt_mapper_computer: StreamFormatMapperComputer[
            L, F
        ] = StdStrFmtMprComputer(),
    ):
        """
        Perform logger configuration using the python's std logger calls.

        ``L`` - python std log logging level type, typically ``int``.

        ``S`` - python std log logging level name type, typically ``str``.

        ``E`` - python std log extended logging level type, typically ``int | str``.

        ``F`` - python std log logging format type, typically ``str``.

        :param level: active logging level.
        :param cmd_name: The command name to register the command logging level to. If ``None`` then the default
            ``COMMAND`` is picked-up and that will be shown on the ``log.cmd()`` call.
        :param stream_fmt_mapper: an output-stream -> log format mapper. Defaults to ``STDERR_ALL_LVL_DIFF_FMT`` if
            ``None`` is supplied. Cannot be used with ``same_fmt_per_lvl``
            and ``stream_set``. Note that ``{}`` denoting an empty ``stream_fmt_mapper`` is accepted and specifies
            the user's intent of not logging to any stream.
        :param same_fmt_per_lvl: Use same log format per logging level. A string argument can be passed to enforce
            that format across all levels. Cannot be provided with ``stream_fmt_mapper``.
        :param stream_set: set of streams to apply level formatting logic to. Cannot be provided with
            ``stream_fmt_mapper``. Note that ``{}`` denoting an empty stream_set is accepted and specifies
            the user's intent of not logging to any stream.
        :param level_name_map: log level - name mapping. This mapping updates the std python logging library's
            registered log levels . Check ``DirectAllLevelLogger.register_levels()`` for more info.
        :param no_warn: do not warn if a supplied level is not registered with the logging library.
        :param propagate: propagate logger records to parent loggers.
        :param handlr_cfgr: The configurator for logger's handlers. Strategy to configure logger's handlers to
            introduce formats on each stream.
        :param stream_fmt_mapper_computer: A computer that computes resulting ``stream_fmt_mapper`` from the supplied
            ``same_fmt_per_lvl`` and ``stream_set`` args.
        """
        self.validate_args(stream_fmt_mapper, stream_set, same_fmt_per_lvl)

        self._level = level
        self.cmd_name = cmd_name
        self.level_name_map = level_name_map
        self.handlr_cfgr = handlr_cfgr
        self.no_warn = no_warn
        self.propagate = propagate
        self.stream_fmt_mapper_computer = stream_fmt_mapper_computer
        if stream_fmt_mapper is not None:  # accepts empty i.e. falsy stream_fmt_mapper
            self.stream_fmt_mapper = stream_fmt_mapper
        else:
            self.stream_fmt_mapper = self.stream_fmt_mapper_computer.compute(
                same_fmt_per_lvl, stream_set
            )
        get_repo().init()

    @override
    def configure(self, logger: logging.Logger) -> DirectAllLevelLogger:
        """
        Configure the std python logger for various formatting quick-hands.

        Examples:

        * Configure with defaults, no errors::

            >>> logger_defaults = StdLoggerConfigurator().configure(logging.getLogger('logger-defaults'))

        * Set ``int`` level::

            >>> logger_int = StdLoggerConfigurator(level=20).configure(logging.getLogger('logger-int'))
            >>> assert logger_int.underlying_logger.level == 20

        * Set digit ``str`` level::

            >>> logger_int_str = StdLoggerConfigurator(level='20').configure(logging.getLogger('logger-int-str'))
            >>> assert logger_int_str.underlying_logger.level == 20

        * Set ``str`` level::

            >>> logger_str = StdLoggerConfigurator(level='FATAL').configure(logging.getLogger('logger-str'))
            >>> assert logger_str.underlying_logger.level == FATAL_LOG_LEVEL

        * ``None`` level sets default `SUCCESS` log level::

            >>> logger_none = (StdLoggerConfigurator(level=None) # noqa
            ...     .configure(logging.getLogger('logger-none')))
            >>> assert logger_none.underlying_logger.level == StdLoggerConfigurator.LOG_LEVEL_DEFAULT_SUCCESS

        * Any other level type raises a ``TypeError``:

          * ``dict`` example:

            >>> logger_dict = (StdLoggerConfigurator(level={}) # noqa
            ...     .configure(logging.getLogger('logger-dict')))
            Traceback (most recent call last):
            TypeError: Wrong level value supplied: '{}', Expected int or str, got dict

          * ``list`` example:

            >>> logger_list = (StdLoggerConfigurator(level=[]) # noqa
            ...     .configure(logging.getLogger('logger-list')))
            Traceback (most recent call last):
            TypeError: Wrong level value supplied: '[]', Expected int or str, got list


        :param logger: std python logger.
        :return: A configured All level logging std python logger.
        """
        stream_fmt_map = self.stream_fmt_mapper
        level = self.level
        levels_to_choose_from: dict[L, S] = DirectAllLevelLogger.register_levels(
            self.level_name_map
        )
        try:
            match level:
                case L():  # typically int
                    int_level = level
                case S():  # typically str
                    int_level = (
                        L(level)
                        if level.isdigit()
                        else logging.getLevelNamesMapping()[level]
                    )
                case None:
                    int_level = StdLoggerConfigurator.LOG_LEVEL_DEFAULT_SUCCESS
                case _:
                    raise TypeError(
                        f"Wrong level value supplied: '{level}', Expected int or str, got "
                        f"{type(level).__name__}"
                    )
        except KeyError:
            if not self.no_warn:
                vt_warn(
                    f"{logger.name}: Undefined log level '{level}'. "
                    f"Choose from {list(levels_to_choose_from.values())}."
                )
                vt_warn(
                    f"{logger.name}: Setting log level to default: "
                    f"'{logging.getLevelName(StdLoggerConfigurator.LOG_LEVEL_DEFAULT_SUCCESS)}'."
                )
            int_level = StdLoggerConfigurator.LOG_LEVEL_DEFAULT_SUCCESS
        logger.setLevel(int_level)
        self.handlr_cfgr.configure(int_level, logger, stream_fmt_map)
        logger.propagate = self.propagate
        possible_level_str: E = logging.getLevelName(level)
        get_repo().index(
            logger.name,
            level=possible_level_str,
            propagate=self.propagate,
            logger="stdlog",
        )
        get_repo().commit()
        return DirectAllLevelLogger(
            DirectAllLevelLoggerImpl(logger), cmd_name=self.cmd_name
        )

    @override
    def set_level(self, new_level: E) -> E:
        orig_level = self.level
        self._level = new_level
        return orig_level

    @override
    @property
    def level(self) -> E:
        return self._level

    @override
    def clone(self, **overrides) -> "StdLoggerConfigurator":
        """
        overrides:
            ``level`` - active logging level.

            ``cmd_name`` - The command name to register the command logging level to. If ``None`` then the default
            ``COMMAND`` is picked-up and that will be shown on the ``log.cmd()`` call.

            ``stream_fmt_mapper`` - an output-stream -> log format mapper. Defaults to ``STDERR_ALL_LVL_DIFF_FMT`` if
            ``None`` is supplied. Cannot be used with ``same_fmt_per_lvl``
            and ``stream_set``. Note that ``{}`` denoting an empty ``stream_fmt_mapper`` is accepted and specifies
            the user's intent of not logging to any stream.

            ``same_fmt_per_lvl`` - Use same log format per logging level. Cannot be provided with
            ``stream_fmt_mapper``.

            ``stream_set`` - set of streams to apply level formatting logic to. Cannot be provided with
            ``stream_fmt_mapper``.Note that ``{}`` denoting an empty stream_set is accepted and specifies the user's
            intent of not logging to any stream.

            ``level_name_map`` - log level - name mapping. This mapping updates the std python logging library's
            registered log levels . Check ``DirectAllLevelLogger.register_levels()`` for more info.

            ``no_warn`` - do not warn if a supplied level is not registered with the logging library.

            ``propagate`` - propagate logger records to parent loggers.

            ``handlr_cfgr`` - The configurator for logger's handlers. Strategy to configure logger's handlers to
            introduce formats on each stream. Check ``logician.stdlog.utils:simple_handlr_cfgr()`` for more info.

            ``stream_fmt_mapper_computer`` - A computer that computes resulting ``stream_fmt_mapper`` from the supplied
            ``same_fmt_per_lvl`` and ``stream_set`` args.
        :return: new ``StdLoggerConfigurator`` with supplied overrides.
        """
        level = overrides.pop("level", self.level)
        cmd_name = overrides.pop("cmd_name", self.cmd_name)
        same_fmt_per_lvl = overrides.pop(
            "same_fmt_per_lvl", StdLoggerConfigurator.FMT_PER_LEVEL_NONE
        )
        stream_set = overrides.pop("stream_set", StdLoggerConfigurator.STREAM_SET_NONE)
        stream_fmt_mapper = overrides.pop("stream_fmt_mapper", None)
        self.validate_args(stream_fmt_mapper, stream_set, same_fmt_per_lvl)
        stream_fmt_mapper = (
            stream_fmt_mapper
            if stream_fmt_mapper is not None
            else self.stream_fmt_mapper
        )
        level_name_map = overrides.pop("level_name_map", self.level_name_map)
        no_warn = overrides.pop("no_warn", self.no_warn)
        propagate = overrides.pop("propagate", StdLoggerConfigurator.PROPAGATE_FALSE)
        handlr_cfgr = overrides.pop("handlr_cfgr", self.handlr_cfgr)
        stream_fmt_mapper_computer = overrides.pop(
            "stream_fmt_mapper_computer", self.stream_fmt_mapper_computer
        )
        if stream_fmt_mapper is not None:
            return StdLoggerConfigurator(
                level=level,
                cmd_name=cmd_name,
                stream_fmt_mapper=stream_fmt_mapper,
                level_name_map=level_name_map,
                no_warn=no_warn,
                propagate=propagate,
                handlr_cfgr=handlr_cfgr,
                stream_fmt_mapper_computer=stream_fmt_mapper_computer,
            )
        else:
            return StdLoggerConfigurator(
                level=level,
                cmd_name=cmd_name,
                stream_set=stream_set,
                same_fmt_per_lvl=same_fmt_per_lvl,
                level_name_map=level_name_map,
                no_warn=no_warn,
                propagate=propagate,
                handlr_cfgr=handlr_cfgr,
                stream_fmt_mapper_computer=stream_fmt_mapper_computer,
            )

    @staticmethod
    def validate_args(
        stream_fmt_mapper: dict[IO, LogLevelFmt[L, F]] | None,
        stream_set: set[IO] | None,
        same_fmt_per_lvl: F | bool | None,
    ):
        """
        :raises ValueError: if  ``stream_fmt_mapper`` is given with ``stream_set`` or
            if  ``stream_fmt_mapper`` is given with ``same_fmt_per_lvl``.
        """
        if stream_fmt_mapper is not None and stream_set is not None:
            raise ValueError(
                errmsg_creator.not_allowed_together("stream_fmt_mapper", "stream_set")
            )
        if stream_fmt_mapper is not None and same_fmt_per_lvl is not None:
            raise ValueError(
                errmsg_creator.not_allowed_together(
                    "stream_fmt_mapper", "same_fmt_per_lvl"
                )
            )


class VQLoggerConfigurator(
    LoggerConfigurator, VQConfigurator[E], HasUnderlyingConfigurator, Protocol
):
    """
    Logger configurator that can decorate other configurators to set their underlying logger levels. This log level is
    to be set according to the supplied verbosity and quietness values.

        ``L`` - python std log logging level type, typically ``int``.

        ``S`` - python std log logging level name type, typically ``str``.

        ``E`` - python std log extended logging level type, typically ``int | str``.

        ``F`` - python std log logging format type, typically ``str``.
    """

    VQ_LEVEL_MAP: VQ_DICT_LITERAL[E] = dict(
        v=logging.INFO,
        vv=logging.DEBUG,
        vvv=TRACE_LOG_LEVEL,
        q=logging.ERROR,
        qq=logging.CRITICAL,
        qqq=FATAL_LOG_LEVEL,
    )
    """
    Default {``verbosity-quietness -> logging-level``} mapping.
    """
    LOG_LEVEL_DEFAULT_SUCCESS: E = DEFAULT_LOG_LEVEL_SUCCESS


class VQSepLoggerConfigurator(VQLoggerConfigurator):
    VQ_LEVEL_MAP_NONE = None
    VQ_SEP_CONF_NONE = None
    LOG_LEVEL_DEFAULT_SUCCESS = VQLoggerConfigurator.LOG_LEVEL_DEFAULT_SUCCESS

    @overload
    def __init__(
        self,
        configurator: LevelLoggerConfigurator[E],
        verbosity: int | None,
        quietness: int | None,
        vq_level_map: VQ_DICT_LITERAL[E] | None = VQ_LEVEL_MAP_NONE,
        vq_sep_configurator: VQSepConfigurator[E] | None = VQ_SEP_CONF_NONE,
        default_log_level: E = LOG_LEVEL_DEFAULT_SUCCESS,
    ): ...

    @overload
    def __init__(
        self,
        configurator: LevelLoggerConfigurator[E],
        verbosity: V_LITERAL | None,
        quietness: Q_LITERAL | None,
        vq_level_map: VQ_DICT_LITERAL[E] | None = VQ_LEVEL_MAP_NONE,
        vq_sep_configurator: VQSepConfigurator[E] | None = VQ_SEP_CONF_NONE,
        default_log_level: E = LOG_LEVEL_DEFAULT_SUCCESS,
    ): ...

    def __init__(
        self,
        configurator: LevelLoggerConfigurator[E],
        verbosity: V_LITERAL | int | None,
        quietness: Q_LITERAL | int | None,
        vq_level_map: VQ_DICT_LITERAL[E] | None = VQ_LEVEL_MAP_NONE,
        vq_sep_configurator: VQSepConfigurator[E] | None = VQ_SEP_CONF_NONE,
        default_log_level: E = LOG_LEVEL_DEFAULT_SUCCESS,
    ):
        """
        A logger configurator that can decorate another logger configurator to accept and infer logging level based on
        ``verbosity`` and ``quietness`` values.

        ``L`` - python std log logging level type, typically ``int``.

        ``S`` - python std log logging level name type, typically ``str``.

        ``E`` - python std log extended logging level type, typically ``int | str``.

        ``F`` - python std log logging format type, typically ``str``.

        Default behavior is::

        - verbosity and quietness are to be supplied separately.
        - default_log_level is returned if both are None or not supplied.
        - if both verbosity and quietness are supplied together then a ValueError is raised.

        Last two behaviors can be altered by choosing a different ``vq_sep_configurator``.

        Examples
        ========

        >>> import warnings

        ``verbosity`` and ``quietness`` cannot be supplied together
        -----------------------------------------------------------
        Warning is issued.
        >>> with warnings.catch_warnings(record=True) as w:
        ...     vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), verbosity='v', quietness='qq')
        ...     assert "verbosity and quietness are not allowed together" in str(w.pop().message)
        >>> assert vq_log.underlying_configurator.level == vq_log.default_log_level

        Default ``VQLoggerConfigurator.VQ_LEVEL_MAP`` is used as ``vq_level_map`` when ``vq_level_map`` is ``None``
        -----------------------------------------------------------------------------------------------------------

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), 'v', None)
        >>> assert vq_log.vq_level_map == VQSepLoggerConfigurator.VQ_LEVEL_MAP

        ``int`` can be supplied for verbosity value
        ------------------------------------------

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), 2, None)
        >>> assert vq_log.verbosity == 'vv'

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), 0, None)
        >>> assert vq_log.verbosity is None

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), None, None)
        >>> assert vq_log.verbosity is None

        ``int`` can be supplied for quietness value
        -------------------------------------------

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), None, 2)
        >>> assert vq_log.quietness == 'qq'

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), None, 0)
        >>> assert vq_log.quietness is None

        >>> vq_log = VQSepLoggerConfigurator(StdLoggerConfigurator(), None, None)
        >>> assert vq_log.quietness is None

        negative ints for verbosity or quietness are not supported
        ----------------------------------------------------------

        >>> VQSepLoggerConfigurator(StdLoggerConfigurator(), -1, None)
        Traceback (most recent call last):
        ValueError: 'verbosity' cannot be negative.

        >>> VQSepLoggerConfigurator(StdLoggerConfigurator(), None, -10)
        Traceback (most recent call last):
        ValueError: 'quietness' cannot be negative.

        over range ints produce warnings
        --------------------------------
        verbosity or quietness > 3 will produce warnings.

        :param configurator: The logger configurator to decorate.
        :param verbosity: verbosity level. Cannot be given with ``quietness``.
        :param quietness: quietness level. Cannot be given with ``verbosity``.
        :param vq_level_map: A user defined {``verbosity-quietness -> logging-level``} mapping can be supplied. Assumes
            ``VQLoggerConfigurator.VQ_LEVEL_MAP`` when omitted or ``None`` is supplied.
        :param vq_sep_configurator: verbosity quietness configurator. Defaults to ``VQSepExclusive``.
        :param default_log_level: log level when none of the verbosity or quietness is supplied.
        """
        self._vq_level_map = (
            vq_level_map if vq_level_map else VQSepLoggerConfigurator.VQ_LEVEL_MAP
        )
        self.vq_sep_configurator = (
            vq_sep_configurator
            if vq_sep_configurator
            else VQSepExclusive(self.vq_level_map, warn_only=True)
        )
        c_verbosity = self.compute_verbosity(
            verbosity, {0: None, 1: "v", 2: "vv", 3: "vvv"}
        )
        c_quietness = self.compute_quietness(
            quietness, {0: None, 1: "q", 2: "qq", 3: "qqq"}
        )
        self.vq_sep_configurator.validate(c_verbosity, c_quietness)
        self._underlying_configurator = configurator
        self.verbosity = c_verbosity
        self.quietness = c_quietness
        self.default_log_level = default_log_level

    @override
    def configure(self, logger: logging.Logger) -> DirectStdAllLevelLogger:
        # TODO: write extensive tests for the
        #  self.underlying_configurator.level or self.default_log_level
        #  logic
        int_level = self.vq_sep_configurator.get_effective_level(
            self.verbosity,
            self.quietness,
            self.underlying_configurator.level or self.default_log_level,
        )
        self.underlying_configurator.set_level(int_level)
        get_repo().index(
            logger.name,
            level=int_level,
            verbosity=self.verbosity,
            quietness=self.quietness,
        )
        return self.underlying_configurator.configure(logger)

    @property
    def vq_level_map(self) -> VQ_DICT_LITERAL[E]:
        return self._vq_level_map

    @override
    @property
    def underlying_configurator(self) -> LevelLoggerConfigurator[E]:
        return self._underlying_configurator  # pragma: no cover

    @override
    def clone(self, **overrides) -> "VQSepLoggerConfigurator":
        """
        overrides:
            ``configurator`` - The logger configurator to decorate.

            ``verbosity`` - verbosity level. Cannot be given with ``quietness``.

            ``quietness`` - quietness level. Cannot be given with ``verbosity``.

            ``vq_level_map`` - A user defined {``verbosity-quietness -> logging-level``} mapping can be supplied.
            Assumes ``VQLoggerConfigurator.VQ_LEVEL_MAP`` when omitted or ``None`` is supplied.

            ``vq_sep_configurator`` - verbosity quietness configurator. Defaults to ``VQSepExclusive``.

            ``default_log_level`` - log level when none of the verbosity or quietness is supplied.
        :return: a new ``VQSepLoggerConfigurator``.
        """
        configurator = overrides.pop("configurator", self.underlying_configurator)
        verbosity = overrides.pop("verbosity", self.verbosity)
        quietness = overrides.pop("quietness", self.quietness)
        vq_level_map = overrides.pop("vq_level_map", self.vq_level_map)
        vq_sep_configurator = overrides.pop(
            "vq_sep_configurator", self.vq_sep_configurator
        )
        default_log_level = overrides.pop("default_log_level", self.default_log_level)
        return VQSepLoggerConfigurator(
            configurator,
            verbosity,
            quietness,
            vq_level_map,
            vq_sep_configurator,
            default_log_level,
        )

    @classmethod
    def compute_verbosity(
        cls, entity: int | V_LITERAL | None, entity_map: dict[int, V_LITERAL | None]
    ) -> V_LITERAL | None:
        return cls._compute_entity(entity, "verbosity", entity_map)

    @classmethod
    def compute_quietness(
        cls, entity: int | Q_LITERAL | None, entity_map: dict[int, Q_LITERAL | None]
    ) -> Q_LITERAL | None:
        return cls._compute_entity(entity, "quietness", entity_map)

    @classmethod
    def _compute_entity(cls, entity, emphasis: str, entity_map: dict):
        if isinstance(entity, int):
            int_entity = int(entity)
            if int_entity < 0:
                raise ValueError(f"'{emphasis}' cannot be negative.")
            max_int_entity = max(entity_map)
            if int_entity > max_int_entity:
                vt_warn(
                    f"Supplied {emphasis}: '{int_entity}' is greater than the max supported "
                    f"{emphasis}: '{max_int_entity}'. Defaulting to max {emphasis}."
                )
                int_entity = max_int_entity
            return entity_map[int_entity]
        else:
            return entity


class VQCommLoggerConfigurator(
    VQLoggerConfigurator, LevelLoggerConfigurator[V_LITERAL | Q_LITERAL | None]
):
    VQ_LEVEL_MAP_NONE = None
    VQ_COMM_CONF_NONE = None
    LOG_LEVEL_DEFAULT_SUCCESS = VQLoggerConfigurator.LOG_LEVEL_DEFAULT_SUCCESS

    def __init__(
        self,
        ver_qui: V_LITERAL | Q_LITERAL | None,
        configurator: LevelLoggerConfigurator[E],
        vq_level_map: VQ_DICT_LITERAL[E] | None = VQ_LEVEL_MAP_NONE,
        vq_comm_configurator: VQCommConfigurator[E] | None = VQ_COMM_CONF_NONE,
        default_log_level: E = LOG_LEVEL_DEFAULT_SUCCESS,
    ):
        """
        A logger configurator that can decorate another logger configurator to accept and infer logging level based on
        ``verbosity`` or ``quietness`` values.

        ``L`` - python std log logging level type, typically ``int``.

        ``S`` - python std log logging level name type, typically ``str``.

        ``E`` - python std log extended logging level type, typically ``int | str``.

        ``F`` - python std log logging format type, typically ``str``.

        Default behavior is::

        - verbosity or quietness is to be supplied in one inclusive argument.
        - default_log_level is returned if both are None or not supplied.

        Last behavior can be altered by choosing a different ``vq_comm_configurator``.

        Examples
        ========

        ``verbosity`` or ``quietness`` to be supplied as one argument.
        --------------------------------------------------------------

        >>> _ = VQCommLoggerConfigurator('qq', StdLoggerConfigurator())

        Default ``VQLoggerConfigurator.VQ_LEVEL_MAP`` is used as ``vq_level_map`` when ``vq_level_map`` is ``None``
        -----------------------------------------------------------------------------------------------------------

        >>> vq_log = VQCommLoggerConfigurator('v', StdLoggerConfigurator())
        >>> assert vq_log.vq_level_map == VQSepLoggerConfigurator.VQ_LEVEL_MAP

        :param configurator: The logger configurator to decorate.
        :param ver_qui: verbosity or quietness level.
        :param vq_level_map: A user defined {``verbosity-quietness -> logging-level``} mapping can be supplied. Assumes
            ``VQLoggerConfigurator.VQ_LEVEL_MAP`` when omitted or ``None`` is supplied.
        :param vq_comm_configurator: verbosity quietness configurator. Defaults to ``VQCommon``.
        :param default_log_level: log level when none of the verbosity or quietness is supplied.
        """
        self._vq_level_map = (
            vq_level_map if vq_level_map else VQCommLoggerConfigurator.VQ_LEVEL_MAP
        )
        self.vq_comm_configurator = (
            vq_comm_configurator
            if vq_comm_configurator
            else VQCommon(self.vq_level_map, warn_only=True)
        )
        self.vq_comm_configurator.validate(ver_qui)
        self._underlying_configurator = configurator
        self.ver_qui = ver_qui
        self.default_log_level = default_log_level

    @override
    def configure(self, logger: logging.Logger) -> DirectStdAllLevelLogger:
        # TODO: write extensive tests for the
        #  self.underlying_configurator.level or self.default_log_level
        #  logic
        int_level = self.vq_comm_configurator.get_effective_level(
            self.ver_qui, self.underlying_configurator.level or self.default_log_level
        )
        get_repo().index(logger.name, vq=self.ver_qui, level=int_level)
        self.underlying_configurator.set_level(int_level)
        return self.underlying_configurator.configure(logger)

    @property
    def vq_level_map(self) -> VQ_DICT_LITERAL[E]:
        return self._vq_level_map

    @property
    def underlying_configurator(
        self,
    ) -> LevelLoggerConfigurator[E]:
        return self._underlying_configurator  # pragma: no cover

    @override
    def set_level(
        self, new_ver_qui: V_LITERAL | Q_LITERAL | None
    ) -> V_LITERAL | Q_LITERAL | None:
        orig_ver_qui = self.ver_qui
        self.ver_qui = new_ver_qui
        return orig_ver_qui

    @override
    @property
    def level(self) -> V_LITERAL | Q_LITERAL | None:
        return self.ver_qui

    @override
    def clone(self, **overrides) -> "VQCommLoggerConfigurator":
        """
        overrides:
            ``configurator`` - The logger configurator to decorate.

            ``ver_qui`` - verbosity or quietness level.

            ``vq_level_map`` - A user defined {``verbosity-quietness -> logging-level``} mapping can be supplied.
            Assumes ``VQLoggerConfigurator.VQ_LEVEL_MAP`` when omitted or ``None`` is supplied.

            ``vq_comm_configurator`` - verbosity quietness configurator. Defaults to ``VQCommon``.

            ``default_log_level`` - log level when none of the verbosity or quietness is supplied.
        :return: a new ``VQCommLoggerConfigurator``.
        """
        configurator = overrides.pop("configurator", self.underlying_configurator)
        ver_qui = overrides.pop("ver_qui", self.ver_qui)
        vq_level_map = overrides.pop("vq_level_map", self.vq_level_map)
        vq_comm_configurator = overrides.pop(
            "vq_comm_configurator", self.vq_comm_configurator
        )
        default_log_level = overrides.pop("default_log_level", self.default_log_level)
        return VQCommLoggerConfigurator(
            ver_qui, configurator, vq_level_map, vq_comm_configurator, default_log_level
        )



#!/usr/bin/env python3
# coding=utf-8

"""
Important utilities for std python logging library.
"""

import logging
from logging import Handler
from typing import IO
from collections import defaultdict

from vt.utils.errors.warnings import vt_warn
from logician.stdlog.constants import LOG_LVL as L, LOG_FMT as F, LOG_STR_LVL as S


def level_name_mapping() -> dict[L, S]:
    """
    :return: level -> name mapping from std lib.
    """
    return {
        level: logging.getLevelName(level)
        for level in sorted(logging.getLevelNamesMapping().values())
    }


class TempSetLevelName:
    def __init__(
        self,
        level: L,
        level_name: S | None,
        reverting_lvl_name: S,
        no_warn: bool = False,
    ):
        """
        Set the log level name temporarily and then revert it back to the ``reverting_lvl_name``.

        :param level: The log level to set name to.
        :param level_name: Level name to set the level to.
        :param reverting_lvl_name: The log level name to revert to when operation finishes.
        :param no_warn: A warning is shown if the supplied ``level_name`` is strip-empty. This warning can be suppressed
            by setting ``no_warn=True``.
        """
        self.level = level
        self.level_name = level_name
        self.reverting_lvl_name = reverting_lvl_name
        self.no_warn = no_warn
        self.original_level_name = logging.getLevelName(level)

    def __enter__(self):
        if self.level_name is not None:
            if self.level_name.strip() == "":
                self.warn_user()
            else:
                logging.addLevelName(self.level, self.level_name)

    def warn_user(self):
        """
        A warning is shown if the supplied ``level_name`` is strip-empty. This warning can be suppressed
            by setting ``no_warn=True`` in ctor.
        """
        if not self.no_warn:
            self._warn_user()

    def _warn_user(self):
        vt_warn(f"Supplied log level name for log level {self.level} is empty.")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.level_name:
            logging.addLevelName(self.level, self.reverting_lvl_name)
        else:
            logging.addLevelName(self.level, self.original_level_name)


def form_stream_handlers_map(logger: logging.Logger) -> dict[IO, list[Handler]]:
    """
    :param logger: the logger whose stream->list[handlers] mapping is to be obtained.
    :return: stream->list[handlers] mapping for the supplied logger.
    """
    stream_handler_map: dict[IO, list[Handler]] = defaultdict(list)
    """
    Map of logger's stream and its handlers.
    """
    # Create a mapping of stream->list[handlers for that stream]
    for handlr in logger.handlers:
        if isinstance(handlr, logging.StreamHandler):
            stream_handler_map[handlr.stream].append(handlr)
    return stream_handler_map


def add_new_formatter(stream: IO, fmt: F) -> logging.StreamHandler:
    """
    Get a handler for ``stream`` with formatter conforming ``fmt`` param.

    >>> import sys

    Examples:

    >>> h = add_new_formatter(sys.stdout, "%(name)s")
    >>> assert h.formatter._fmt == "%(name)s"   # type: ignore[attr-defined]

    :param stream: A new ``logging.StreamHandler`` will be created for this stream.
    :param fmt: A formatter conforming to ``fmt`` will be set for ``stream``.
    :return: a new handler with the formatter set to ``fmt``.
    """
    _handlr = logging.StreamHandler(stream=stream)  # type: ignore[arg-type]
    _handlr.setFormatter(logging.Formatter(fmt=fmt))
    return _handlr



#!/usr/bin/env python3
# coding=utf-8

"""
Extract and showcase details about a program's logger configurators.
"""

import os
import shlex
import sys
import argparse
import subprocess
import tempfile
from collections import defaultdict
from pathlib import Path
from typing import Any
from logician.errors import (
    LogicianExitingException,
    LogicianCmdException,
    LogicianCmdNotFoundError,
    LogicianException,
)

from vt.utils.commons.commons.strings import generate_random_string

from vt.utils.errors.error_specs import ERR_INVALID_USAGE, ERR_CMD_NOT_FOUND

from logician.constants import LGCN_MAIN_CMD_NAME, LGCN_INFO_FP_ENV_VAR

# TODO: add extensive examples in the README. Better yet, create a whole separate file/section for examples.

examples = """
Examples:

Just see the logger names of a command that uses logician:

    # use `lgcn command [command...]` syntax    
    $ lgcn prog-using-logician another-prog-using-lgcn prog-using-lgcn-but-no-loggers prog-not-using-lgcn
    prog-using-logician: ['prog', 'prog.child', 'sibling'] # list of logger names
    ano-prog-using-lgcn: ['r', 'r.c', 's', 's.c'] # list of logger names
    prog-using-lgcn-but-no-loggers: [] # no logger names visible
    # prog-not-using-lgcn is not visible.
    
Use the long listing format:

    # using `lgcn -l command [command...]` syntax
    # uses default `{cmd}\t{name}\t{level}\t{vq-support}\t{env-support}` format
    $ lgcn -l prog-1 prog-2 prog-3
    command             logger              level               vq-support          env-support         
    ----------------------------------------------------------------------------------------------------
    prog-1              ro                  Level 14            False                ['IN_ENV', 'IN.ENV']
    prog-1              ro.or               Level 14            False               False
    prog-1              r                   Level 32            True                ['ENV']
    prog-1              r.c                 Level 32            True                ['ENV.C', 'ENV_C', 'ENV']
    prog-1              c.r                 Level 12            False               False
    prog-2              .ano                WARNING             False               {}
    
    # prog-1 and prog-2 use logician, prog-3 doesn't.
    # vq-support denotes whether the env-vars can take 'v', 'vv', 'vvv' for verbosity levels and 
    # 'q', 'qq', 'qqq' for quietness level settings.
    
List all the log env-vars supported by programs:

    # using `lgcn -e command [command...]` syntax
    $ lgcn -e prog-1 prog-2 prog-3
    prog-1: ['IN_ENV', 'IN.ENV', 'ENV', 'ENV.C', 'ENV_C', 'ENV']
    prog-2: []
    
    # prog-1 uses logician and exposes ['IN_ENV', 'IN.ENV', 'ENV', 'ENV.C', 'ENV_C', 'ENV'] env-vars
    # prog-2 uses logician but does not export any env-vars.
    # prog-3 does not use logician
    
Use long listing format with env-vars:

    # using `lgcn -l -e command [command...]` syntax
    # uses default `{cmd}\t{name}\t{level}\t{vq-support}\t{env-support}` format
    $ lgcn -le prog-1 prog-2 prog-3
    command             logger              level               vq-support          env-support         
    ----------------------------------------------------------------------------------------------------
    prog-1              ro                  Level 14            False                ['IN_ENV', 'IN.ENV']
    prog-1              ro.or               Level 14            False               []
    prog-1              r                   Level 32            True                ['ENV']
    prog-1              r.c                 Level 32            True                ['ENV.C', 'ENV_C', 'ENV']
    prog-1              c.r                 Level 12            False               []
    prog-2              .ano                WARNING             False               []

    # prog-1 uses logician and exposes ['IN_ENV', 'IN.ENV', 'ENV', 'ENV.C', 'ENV_C', 'ENV'] env-vars
    # prog-2 uses logician but does not export any env-vars.
    # prog-3 does not use logician
    
"""

CONST_FMT = "{cmd}\t{name}\t{level}\t{vq-support}\t{env-support}"


def main(*commands: str) -> dict[str, dict[str, dict[str, Any]]]:
    """
    Assumes that each command supports -h option.

    :param commands: commands to get the logger configurator details for.
    :return: a dictionary of command and their individual logger-configurator properties.
    :raises VTCmdException: if error in running ``<<supplied-command>> --help`` for each command.
    """
    cmd_det_dict: dict[str, dict[str, dict[str, Any]]] = dict()  # pragma: no cover
    env_fp: Path = Path(
        tempfile.gettempdir(),
        f".0-LGCN-{'-'.join(commands)}-{generate_random_string()}.json",
    )
    os.environ[LGCN_INFO_FP_ENV_VAR] = str(env_fp)
    from logician._repo import get_repo

    get_repo().init()
    for command in commands:
        env_fp.write_text("")  # quick and dirty way to reinitialise the repo files.
        # This logic must somehow be handled into the repo provider itself.
        # TODO: If the env_fp.write("") statement is removed then, the file present in the /tmp dir formed by env_fp
        #  remains with the previous command's logger-configurator details. This is erroneous as then any command which
        #  does not use logger-configurators actually gets the configurator details which are already stored in the
        #  env_fp file from the previous command run.
        #  To get a feel of what this error entails, run these two commands:
        #  lgcn grep <a-command-that-uses-logician> # this will output grep: {}, i.e. grep has no logger-configurators
        #  lgcn <a-command-that-uses-logician> grep # this will output grep: <details-of-logger-configurators-of-the-second-program>,
        #  i.e. grep is incorrectly shown the logger configurator values of other programs as the env_fp file still
        #  stores those details from the previous run.
        #  FIX THIS!
        try:
            subprocess.run(
                [*shlex.split(command), "--help"], capture_output=True, check=True
            )
        except FileNotFoundError as f:
            raise LogicianCmdNotFoundError(
                command=shlex.split(command),
                file_not_found_error=f,
                exit_code=ERR_CMD_NOT_FOUND,
            ) from f
        except subprocess.CalledProcessError as e:
            raise LogicianCmdException(
                f"Command failed: {e.cmd}",
                f"Stderr: {e.stderr}",
                f"Stdout: {e.stdout}",
                called_process_error=e,
                exit_code=e.returncode,
            ) from e
        get_repo().reload()
        cmd_det_dict[command] = get_repo().read_all()
    return cmd_det_dict


def cli(args: list[str]) -> argparse.Namespace:
    """
    Examples:

    >>> cli(["cmd1"])
    Namespace(command=['cmd1'], ls=False, fmt=None, env_list=False)

    >>> cli(["cmd1", "cmd2"])
    Namespace(command=['cmd1', 'cmd2'], ls=False, fmt=None, env_list=False)

    >>> cli(["cmd1", "cmd2", "-l"])
    Namespace(command=['cmd1', 'cmd2'], ls=True, fmt=None, env_list=False)

    >>> cli(["cmd1", "cmd2", "-le"])
    Namespace(command=['cmd1', 'cmd2'], ls=True, fmt=None, env_list=True)

    >>> cli([])
    Traceback (most recent call last):
    ...
    SystemExit: 2

    >>> cli(['cmd1', '--fmt'])
    Traceback (most recent call last):
    ...
    SystemExit: 2

    :param args: arguments to the ``lgcn`` CLI.
    :return: Calculated ``argparse.Namespace`` from ``lgcn`` CLI.
    """
    parser = argparse.ArgumentParser(
        LGCN_MAIN_CMD_NAME,
        description=__doc__,
        add_help=False,
    )
    parser.add_argument(
        "command",
        help="get logger-configurator details of these commands. Assumes that all of these commands "
        "support the --help CLI option.",
        nargs="+",
    )
    # Helpers group
    helper_group = parser.add_argument_group("helps", "Get help regarding logician")
    helper_group.add_argument("-h", help="Show compact help and exit.", action="help")
    helper_group.add_argument(
        "--help", help="Show extended help and exit.", action="help"
    )
    lister_group = parser.add_argument_group(
        "listing", "options related to listing details about the logger-configurators"
    )
    lister_group.add_argument(
        "-l", "--list", action="store_true", help="Use long listing format.", dest="ls"
    )
    lister_group.add_argument(
        "--fmt",
        "--format",
        const=CONST_FMT,
        nargs="?",
        help="""Can only be used with -l option. 
        Print formatted information about logger-configurators. 
        More headers, like, {lib}, {stream}, {no-of-handlers}, ...etc are available. Check documentation.""",
        dest="fmt",
    )
    parser.add_argument(
        "-e",
        "--env-list",
        action="store_true",
        help="Get supported environment variables list.",
    )

    if "--help" in args:
        print(parser.format_help() + examples)
        sys.exit()

    namespace: argparse.Namespace = parser.parse_args(args)
    if namespace.fmt and not namespace.ls:
        parser.error("--format is only allowed with --list")
    return namespace


def main_view(
    info_dict: dict[str, dict[str, dict[str, Any]]],
    ls: bool,
    env_list: bool,
    fmt: str | None = None,
):
    """
    View that will print info about ``info_dict`` on ``stdout`` according to the required formats.

    Examples:

    >>> main_view(dict(), False, True, "{name}")
    Traceback (most recent call last):
    ...
    logician.errors.LogicianExitingException: ValueError: fmt cannot be used when ls is Falsy.

    >>> main_view(dict(), None, # type: ignore[arg-type]
    ...             True, "{name}")
    Traceback (most recent call last):
    ...
    logician.errors.LogicianExitingException: ValueError: fmt cannot be used when ls is Falsy.

    :param info_dict: mappings of commands and their individual logger configurator details.
    :param ls: Use long listing format
    :param env_list: show supported env vars.
    :param fmt: list in supplied formats. Can only be used when ``ls`` is True.
    """
    if fmt is not None and not ls:
        errmsg = "fmt cannot be used when ls is Falsy."
        raise LogicianExitingException(
            errmsg, exit_code=ERR_INVALID_USAGE
        ) from ValueError(errmsg)

    el_det: dict[str, list[str]] = defaultdict(
        list
    )  # env list details. cmd -> env-list mapping
    ls_det: dict = defaultdict(
        dict
    )  # ls default details. cmd -> {name, level, vq_support, env_support}

    # Prepare env-vars
    for cmd in info_dict:
        for lgr in info_dict[cmd]:
            # Run for all the registered loggers.
            # makes sure that programs not using logician do not get registered here.
            if "env_list" in info_dict[cmd][lgr]:
                el_det[cmd].extend(info_dict[cmd][lgr]["env_list"])
            else:
                if not el_det[cmd]:
                    el_det[cmd] = []

    # Prepare list in the predetermined fmt
    for cmd in info_dict:
        for lgr in info_dict[cmd]:
            ls_det[cmd][lgr] = defaultdict(dict)
            if "vq" in info_dict[cmd][lgr]:
                ls_det[cmd][lgr]["vq_support"] = True
            else:
                ls_det[cmd][lgr]["vq_support"] = False

            if cmd in el_det:
                if "env_list" in info_dict[cmd][lgr]:
                    if env_list:
                        ls_det[cmd][lgr]["env_support"] = info_dict[cmd][lgr][
                            "env_list"
                        ]
                    else:
                        ls_det[cmd][lgr]["env_support"] = True
                else:
                    if env_list:
                        ls_det[cmd][lgr]["env_support"] = []
                    else:
                        ls_det[cmd][lgr]["env_support"] = False
            ls_det[cmd][lgr]["level"] = info_dict[cmd][lgr]["level"]

    if ls:
        # Only print list in the predetermined fmt
        frmt = "{:<20}" * 5  # 5 columns
        print(frmt.format("command", "logger", "level", "vq-support", "env-support"))
        print("-" * 20 * 5)
        for cmd, lgr in ls_det.items():
            [
                print(
                    frmt.format(
                        cmd,
                        _c,
                        _l["level"],
                        str(_l["vq_support"]),
                        str(_l["env_support"]),
                    )
                )
                for _c, _l in lgr.items()
            ]
        return

    if env_list:
        # Only print env-list per command
        [print(f"{_c}: {_l}") for _c, _l in el_det.items()]
        return

    # Simply print logger names per command
    ln_det = defaultdict(list)  # logger names per command
    for cmd in info_dict:
        for lgr in info_dict[cmd]:
            ln_det[cmd].append(lgr)
    [print(f"{_c}: {_l}") for _c, _l in ln_det.items()]


def main_cli(args: list[str] | None = None):
    """
    Main CLI, this runs:

    - Just the CLI to get namespace.
    - Main logic to get a mapping of command and its individual logger-configurators.
    - Print the command -> logger-configurator mappings according to user supplied options.

    :param args: CLI args to ``lgcn``.
    """
    try:
        args = args if args else sys.argv[1:]
        namespace: argparse.Namespace = cli(args)
        info_dict: dict[str, dict[str, dict[str, Any]]] = main(
            *namespace.command,
        )
        main_view(
            info_dict,
            ls=namespace.ls,
            env_list=namespace.env_list,
            fmt=namespace.fmt,
        )
    except LogicianExitingException as _le:
        print(_le, file=sys.stderr)
        sys.exit(_le.exit_code)
    except LogicianException as _l:
        print(_l, file=sys.stderr)


if __name__ == "__main__":
    main_cli()



#!/usr/bin/env python3
# coding=utf-8

"""
Logging related library.

This library is built to let other logging projects extend it and implement their own versions of various log levels.
Only log levels are of essence in this library as most of the logging concerns log levels and not the actual logger
setup.

Aim is to let users extend this library to fit their various log levels and logging goals and not think twice before
replacing one logger implementation with another.

Extending libraries are to be designed thinking that the underlying logger will be supplied by the client/caller
at the time of logician logger configurator creation and hence client can configure the logger as they please before
supplying the logger class to perform delegation onto by this library.
"""

from logging import Logger


# region base re-exports
from logician.base import AllLevelLogger as AllLevelLogger
from logician.base import MinLogProtocol as MinLogProtocol
# endregion

# region std-log re-exports
from logician.stdlog import StdLogProtocol as StdLogProtocol
from logician.stdlog import StdLevelLogger as StdLevelLogger
from logician.stdlog import StdProtocolAllLevelLogger as StdProtocolAllLevelLogger
from logician.stdlog import BaseDirectStdAllLevelLogger as BaseDirectStdAllLevelLogger
from logician.stdlog import DirectAllLevelLogger as DirectAllLevelLogger
from logician.stdlog import DirectStdAllLevelLogger as DirectStdAllLevelLogger
# endregion

from logician.stdlog.all_levels_impl import DirectAllLevelLoggerImpl as _DALImpl

from logician.constants import LGCN_ALL_LOG_ENV_VAR as LGCN_ALL_LOG_ENV_VAR

from logician.utils import command_or_file as command_or_file

from vt.utils.errors.error_specs import ErrorMsgFormer as ErrorMsgFormer

errmsg_creator = ErrorMsgFormer
"""
Create formatted error messages using this global instance.

To get a local instance use ``errmsg_creator.clone_with(...)``.
"""


def get_direct_all_level_logger(logger: Logger) -> DirectStdAllLevelLogger:
    """
    Simple logger configurator to directly configure a std logger to logician standards.

    :param logger: python std logger.
    :return: logician configured logger.
    """
    return DirectAllLevelLogger(_DALImpl(logger))
